# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for specs commands."""


import importlib
import os
import shutil
from pathlib import Path
from typing import Tuple

from autonomy.analyse.abci.app_spec import DFA

from tests.conftest import ROOT_DIR
from tests.test_autonomy.test_cli.base import BaseCliTest


class TestGenerateSpecs(BaseCliTest):
    """Test generate-app-specs"""

    cli_options: Tuple[str, ...] = ("analyse", "abci", "generate-app-specs")
    skill_path = Path("packages", "valory", "skills", "hello_world_abci", "rounds")
    app_name = "HelloWorldAbciApp"

    cls_name: str
    dfa: DFA

    @classmethod
    def setup(cls) -> None:
        """Setup class."""
        super().setup()

        module_name = ".".join(cls.skill_path.parts)
        module = importlib.import_module(module_name)
        cls.cls_name = ".".join([module_name, cls.app_name])

        abci_app_cls = getattr(module, cls.app_name)
        cls.dfa = DFA.abci_to_dfa(abci_app_cls, cls.cls_name)

    def get_expected_output(self, output_format: str) -> str:
        """Get expected output."""

        temp_file = self.t / "temp"
        self.dfa.dump(temp_file, output_format)

        return temp_file.read_text(encoding="utf-8")

    def _run_test(self, output_format: str) -> None:
        """Run test for given output format type."""

        output_file = self.t / "fsm"
        result = self.run_cli((self.cls_name, str(output_file), f"--{output_format}"))

        assert result.exit_code == 0
        assert output_file.read_text() == self.get_expected_output(output_format)

    def test_generate_yaml(
        self,
    ) -> None:
        """Run tests."""

        self._run_test(DFA.OutputFormats.YAML)

    def test_generate_json(
        self,
    ) -> None:
        """Run tests."""

        self._run_test(DFA.OutputFormats.JSON)

    def test_generate_mermaid(
        self,
    ) -> None:
        """Run tests."""

        self._run_test(DFA.OutputFormats.MERMAID)


class TestCheckSpecs(BaseCliTest):
    """Test `check-app-specs` command."""

    cli_options: Tuple[str, ...] = ("analyse", "abci", "check-app-specs")
    skill_path = Path("packages", "valory", "skills", "hello_world_abci", "rounds")
    app_name = "HelloWorldAbciApp"

    specification_path: Path
    cls_name: str
    dfa: DFA

    @classmethod
    def setup(cls) -> None:
        """Setup class."""
        super().setup()

        module_name = ".".join(cls.skill_path.parts)
        module = importlib.import_module(module_name)
        cls.cls_name = ".".join([module_name, cls.app_name])

        abci_app_cls = getattr(module, cls.app_name)
        cls.dfa = DFA.abci_to_dfa(abci_app_cls, cls.cls_name)

        cls.specification_path = cls.skill_path.parent / "fsm_specification.yaml"
        shutil.copytree(ROOT_DIR / "packages", cls.t / "packages")
        os.chdir(cls.t)

    def _corrupt_spec_file(
        self,
    ) -> None:
        """Corrupt spec file to fail the check."""
        content = self.specification_path.read_text()
        content = content.replace(
            "(SelectKeeperRound, ROUND_TIMEOUT): RegistrationRound\n", ""
        )
        self.specification_path.write_text(content)

    def _fix_corrupt_file(
        self,
    ) -> None:
        """Fix corrupt file."""
        self.specification_path.write_text(
            Path(*ROOT_DIR.absolute().parts, *self.specification_path.parts).read_text()
        )

    def test_one_pass(
        self,
    ) -> None:
        """Test with one class."""
        self._fix_corrupt_file()
        result = self.run_cli(
            ("--app-class", self.cls_name, "--infile", str(self.specification_path))
        )

        assert result.exit_code == 0
        assert (
            result.output
            == "Checking : packages.valory.skills.hello_world_abci.rounds.HelloWorldAbciApp\nCheck successful\n"
        )

    def test_one_fail(
        self,
    ) -> None:
        """Test with one class failing."""
        self._corrupt_spec_file()
        result = self.run_cli(
            ("--app-class", self.cls_name, "--infile", str(self.specification_path))
        )

        assert result.exit_code == 1
        assert (
            result.output
            == "Checking : packages.valory.skills.hello_world_abci.rounds.HelloWorldAbciApp\nCheck failed.\n"
        )

    def test_check_all(
        self,
    ) -> None:
        """Test --check-all flag."""

        self._fix_corrupt_file()
        result = self.run_cli(
            (
                "--check-all",
                "--packages-dir",
                str(self.t / "packages"),
            )
        )

        assert result.exit_code == 0
        assert "Check successful." in result.output

    def test_check_all_fail(
        self,
    ) -> None:
        """Test --check-all flag."""

        self._corrupt_spec_file()
        result = self.run_cli(
            (
                "--check-all",
                "--packages-dir",
                str(self.t / "packages"),
            )
        )

        assert result.exit_code == 1
        assert (
            "Specifications did not match for following definitions." in result.output
        )

    @classmethod
    def teardown(cls) -> None:
        """Teardown method."""
        os.chdir(cls.cwd)
        super().teardown()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the data classes for the price estimation ABCI application."""

import statistics
from enum import Enum
from typing import Callable, Dict, Set, Type, cast

from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppTransitionFunction,
    AbstractRound,
    AppState,
    BaseSynchronizedData,
    CollectDifferentUntilThresholdRound,
    CollectSameUntilThresholdRound,
    DegenerateRound,
)
from packages.valory.skills.price_estimation_abci.payloads import (
    EstimatePayload,
    ObservationPayload,
    TransactionHashPayload,
)


class Event(Enum):
    """Event enumeration for the price estimation demo."""

    DONE = "done"
    NONE = "none"
    ROUND_TIMEOUT = "round_timeout"
    NO_MAJORITY = "no_majority"


class SynchronizedData(BaseSynchronizedData):
    """
    Class to represent the synchronized data.

    This data is replicated by the tendermint application.
    """

    _aggregator_method: Callable
    _aggregator_methods: Dict[str, Callable] = {
        "mean": statistics.mean,
        "median": statistics.median,
        "mode": statistics.mode,
    }

    def set_aggregator_method(self, aggregator_method: str) -> None:
        """Set aggregator method."""
        self._aggregator_method = self._aggregator_methods.get(  # type: ignore
            aggregator_method, statistics.median
        )

    @property
    def safe_contract_address(self) -> str:
        """Get the safe contract address."""
        return cast(str, self.db.get_strict("safe_contract_address"))

    @property
    def oracle_contract_address(self) -> str:
        """Get the oracle contract address."""
        return cast(str, self.db.get_strict("oracle_contract_address"))

    @property
    def estimate(self) -> float:
        """Get the estimate."""
        observations = [
            value.observation for value in self.participant_to_observations.values()
        ]
        return self._aggregator_method(observations)

    @property
    def most_voted_estimate(self) -> float:
        """Get the most_voted_estimate."""
        return cast(float, self.db.get_strict("most_voted_estimate"))

    @property
    def most_voted_tx_hash(self) -> float:
        """Get the most_voted_tx_hash."""
        return cast(float, self.db.get_strict("most_voted_tx_hash"))

    @property
    def participant_to_observations(self) -> Dict:
        """Get the participant_to_observations."""
        return cast(Dict, self.db.get_strict("participant_to_observations"))

    @property
    def participant_to_estimate(self) -> Dict:
        """Get the participant_to_estimate."""
        return cast(Dict, self.db.get_strict("participant_to_estimate"))


class CollectObservationRound(CollectDifferentUntilThresholdRound):
    """A round in which agents collect observations"""

    round_id = "collect_observation"
    allowed_tx_type = ObservationPayload.transaction_type
    payload_attribute = "observation"
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    no_majority_event = Event.NO_MAJORITY
    selection_key = "participant"
    collection_key = "participant_to_observations"


class EstimateConsensusRound(CollectSameUntilThresholdRound):
    """A round in which agents reach consensus on an estimate"""

    round_id = "estimate_consensus"
    allowed_tx_type = EstimatePayload.transaction_type
    payload_attribute = "estimate"
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    none_event = Event.NONE
    no_majority_event = Event.NO_MAJORITY
    collection_key = "participant_to_estimate"
    selection_key = "most_voted_estimate"


class TxHashRound(CollectSameUntilThresholdRound):
    """A round in which agents compute the transaction hash"""

    round_id = "tx_hash"
    allowed_tx_type = TransactionHashPayload.transaction_type
    payload_attribute = "tx_hash"
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    none_event = Event.NONE
    no_majority_event = Event.NO_MAJORITY
    collection_key = "participant_to_tx_hash"
    selection_key = "most_voted_tx_hash"


class FinishedPriceAggregationRound(DegenerateRound):
    """A round that represents price aggregation has finished"""

    round_id = "finished_price_aggregation"


class PriceAggregationAbciApp(AbciApp[Event]):
    """PriceAggregationAbciApp

    Initial round: CollectObservationRound

    Initial states: {CollectObservationRound}

    Transition states:
        0. CollectObservationRound
            - done: 1.
            - round timeout: 0.
            - no majority: 0.
        1. EstimateConsensusRound
            - done: 2.
            - round timeout: 0.
            - no majority: 0.
        2. TxHashRound
            - done: 3.
            - none: 0.
            - round timeout: 0.
            - no majority: 0.
        3. FinishedPriceAggregationRound

    Final states: {FinishedPriceAggregationRound}

    Timeouts:
        round timeout: 30.0
    """

    initial_round_cls: Type[AbstractRound] = CollectObservationRound
    transition_function: AbciAppTransitionFunction = {
        CollectObservationRound: {
            Event.DONE: EstimateConsensusRound,
            Event.ROUND_TIMEOUT: CollectObservationRound,
            Event.NO_MAJORITY: CollectObservationRound,
        },
        EstimateConsensusRound: {
            Event.DONE: TxHashRound,
            Event.ROUND_TIMEOUT: CollectObservationRound,
            Event.NO_MAJORITY: CollectObservationRound,
        },
        TxHashRound: {
            Event.DONE: FinishedPriceAggregationRound,
            Event.NONE: CollectObservationRound,
            Event.ROUND_TIMEOUT: CollectObservationRound,
            Event.NO_MAJORITY: CollectObservationRound,
        },
        FinishedPriceAggregationRound: {},
    }
    final_states: Set[AppState] = {FinishedPriceAggregationRound}
    event_to_timeout: Dict[Event, float] = {
        Event.ROUND_TIMEOUT: 30.0,
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the behaviours for the 'abci' skill."""
import json
from typing import Generator, Set, Type

from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.registration_abci.payloads import RegistrationPayload
from packages.valory.skills.registration_abci.rounds import (
    AgentRegistrationAbciApp,
    RegistrationRound,
    RegistrationStartupRound,
)


class RegistrationBaseBehaviour(BaseBehaviour):
    """Agent registration to the FSM App."""

    def async_act(self) -> Generator:
        """
        Do the action.

        Steps:
        - Build a registration transaction.
        - Send the transaction and wait for it to be mined.
        - Wait until ABCI application transitions to the next round.
        - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            initialisation = (
                json.dumps(self.synchronized_data.db.initial_data, sort_keys=True)
                if self.synchronized_data.db.initial_data != {}
                else None
            )
            payload = RegistrationPayload(
                self.context.agent_address, initialisation=initialisation
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class RegistrationStartupBehaviour(RegistrationBaseBehaviour):
    """Agent registration to the FSM App."""

    behaviour_id = "registration_startup"
    matching_round = RegistrationStartupRound


class RegistrationBehaviour(RegistrationBaseBehaviour):
    """Agent registration to the FSM App."""

    behaviour_id = "registration"
    matching_round = RegistrationRound


class AgentRegistrationRoundBehaviour(AbstractRoundBehaviour):
    """This behaviour manages the consensus stages for the registration."""

    initial_behaviour_cls = RegistrationStartupBehaviour
    abci_app_cls = AgentRegistrationAbciApp  # type: ignore
    behaviours: Set[Type[BaseBehaviour]] = {  # type: ignore
        RegistrationBehaviour,  # type: ignore
        RegistrationStartupBehaviour,  # type: ignore
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for docstring generator."""


import os
import shutil
from pathlib import Path
from typing import Tuple

import pytest

from tests.conftest import ROOT_DIR
from tests.test_autonomy.test_cli.base import BaseCliTest


@pytest.mark.skip
class TestDocstrings(BaseCliTest):
    """Test `autonomy analyse abci docstrings`."""

    rounds_file_original: Path
    rounds_file_temp: Path
    cli_options: Tuple[str, ...] = ("analyse", "abci", "docstrings")
    rounds_file = Path("packages", "valory", "skills", "hello_world_abci", "rounds.py")

    @classmethod
    def setup(
        cls,
    ) -> None:
        """Setup class."""
        super().setup()

        shutil.copytree(ROOT_DIR / "packages", cls.t / "packages")
        os.chdir(cls.t)

        cls.rounds_file_original = Path(*ROOT_DIR.parts, *cls.rounds_file.parts)
        cls.rounds_file_temp = Path(*cls.t.parts, *cls.rounds_file.parts)

    def _get_expected_output(
        self,
    ) -> str:
        """Return expected output."""
        expcted_output = ""
        for file in sorted(self.t.glob("packages/*/skills/*/rounds.py")):
            expcted_output += f"Processing: {file.relative_to(self.t)}\n"
        return expcted_output

    def _corrupt_round_file(
        self,
    ) -> None:
        """Corrupt a round file."""

        string_to_replace = """3. ResetAndPauseRound
            - done: 1.
            - reset timeout: 0.
            - no majority: 0.\n"""

        rounds_file_content = self.rounds_file_original.read_text()
        rounds_file_content = rounds_file_content.replace(string_to_replace, "")
        self.rounds_file_temp.write_text(rounds_file_content)

    def _fix_round_file(
        self,
    ) -> None:
        """Fix the corrupted round file."""
        self.rounds_file_temp.write_text(self.rounds_file_original.read_text())

    def test_docstring_check(
        self,
    ) -> None:
        """Test after check"""

        packages_dir = self.t.relative_to(Path.cwd()) / "packages"
        self.run_cli((str(packages_dir),))
        result = self.run_cli((str(packages_dir), "--check"))

        expcted_output = self._get_expected_output()
        expcted_output += "No update needed.\n"
        assert result.exit_code == 0
        assert result.output == expcted_output

    def test_docstring_check_fail(
        self,
    ) -> None:
        """Test after check"""

        packages_dir = self.t.relative_to(Path.cwd()) / "packages"
        self.run_cli((str(packages_dir),))

        self._corrupt_round_file()
        result = self.run_cli((str(packages_dir), "--check"))
        self._fix_round_file()

        expcted_output = self._get_expected_output()
        expcted_output += (
            "Error: Following files needs updating.\n"
            "\n"
            "packages/valory/skills/hello_world_abci/rounds.py\n"
        )
        assert result.exit_code == 1
        assert result.output == expcted_output

    def test_fix_docstring(
        self,
    ) -> None:
        """Test after check"""

        packages_dir = self.t.relative_to(Path.cwd()) / "packages"
        self.run_cli((str(packages_dir),))

        self._corrupt_round_file()
        result = self.run_cli((str(packages_dir),))

        expcted_output = self._get_expected_output()
        expcted_output += (
            "\nUpdated following files.\n"
            "\npackages/valory/skills/hello_world_abci/rounds.py\n"
        )
        assert result.exit_code == 0
        assert result.output == expcted_output
        assert (
            self.rounds_file_original.read_text() == self.rounds_file_temp.read_text()
        )

    @classmethod
    def teardown(cls) -> None:
        """Teardown class."""
        os.chdir(ROOT_DIR)
        super().teardown()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains utility functions for the 'abstract_round_abci' skill."""

from hashlib import sha256
from typing import Dict, Optional, Tuple, Union, cast

from eth_typing.bls import BLSPubkey, BLSSignature
from py_ecc.bls import G2Basic as bls  # type: ignore


MAX_UINT64 = 2 ** 64 - 1


class VerifyDrand:  # pylint: disable=too-few-public-methods
    """
    Tool to verify Randomness retrieved from various external APIs.

    The ciphersuite used is BLS_SIG_BLS12381G2_XMD:SHA-256_SSWU_RO_NUL_

    https://drand.love/docs/specification/#cryptographic-specification
    https://github.com/ethereum/py_ecc
    """

    @classmethod
    def _int_to_bytes_big(cls, value: int) -> bytes:
        """Convert int to bytes."""
        if value < 0 or value > MAX_UINT64:
            raise ValueError(
                "VerifyDrand can only handle positive numbers representable with 8 bytes"
            )
        return int.to_bytes(value, 8, byteorder="big", signed=False)

    @classmethod
    def _verify_randomness_hash(cls, randomness: bytes, signature: bytes) -> bool:
        """Verify randomness hash."""
        return sha256(signature).digest() == randomness

    @classmethod
    def _verify_signature(
        cls,
        pubkey: Union[BLSPubkey, bytes],
        message: bytes,
        signature: Union[BLSSignature, bytes],
    ) -> bool:
        """Verify randomness signature."""
        return bls.Verify(
            cast(BLSPubkey, pubkey), message, cast(BLSSignature, signature)
        )

    def verify(self, data: Dict, pubkey: str) -> Tuple[bool, Optional[str]]:
        """
        Verify drand value retried from external APIs.

        :param data: dictionary containing drand parameters.
        :param pubkey: league of entropy public key
                       https://drand.love/developer/http-api/#public-endpoints
        :returns: bool, error message
        """

        encoded_pubkey = bytes.fromhex(pubkey)
        try:
            randomness = data["randomness"]
            signature = data["signature"]
            round_value = int(data["round"])
        except KeyError as e:
            return False, f"DRAND dict is missing value for {e}"

        previous_signature = data.pop("previous_signature", "")
        encoded_randomness = bytes.fromhex(randomness)
        encoded_signature = bytes.fromhex(signature)
        int_encoded_round = self._int_to_bytes_big(round_value)
        encoded_previous_signature = bytes.fromhex(previous_signature)

        if not self._verify_randomness_hash(encoded_randomness, encoded_signature):
            return False, "Failed randomness hash check."

        msg_b = encoded_previous_signature + int_encoded_round
        msg_hash_b = sha256(msg_b).digest()

        if not self._verify_signature(encoded_pubkey, msg_hash_b, encoded_signature):
            return False, "Failed bls.Verify check."

        return True, None



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the liquidity rebalancing ABCI application."""

from packages.valory.skills.abstract_round_abci.abci_app_chain import (
    AbciAppTransitionMapping,
    chain,
)
from packages.valory.skills.liquidity_rebalancing_abci.rounds import (
    FinishedEnterPoolTransactionHashRound,
    FinishedExitPoolTransactionHashRound,
    FinishedSwapBackTransactionHashRound,
    LiquidityRebalancingAbciApp,
    StrategyEvaluationRound,
)
from packages.valory.skills.registration_abci.rounds import (
    AgentRegistrationAbciApp,
    FinishedRegistrationFFWRound,
    FinishedRegistrationRound,
    RegistrationRound,
)
from packages.valory.skills.reset_pause_abci.rounds import (
    FinishedResetAndPauseErrorRound,
    FinishedResetAndPauseRound,
    ResetAndPauseRound,
    ResetPauseABCIApp,
)
from packages.valory.skills.safe_deployment_abci.rounds import (
    FinishedSafeRound,
    RandomnessSafeRound,
    SafeDeploymentAbciApp,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    FailedRound,
    FinishedTransactionSubmissionRound,
    RandomnessTransactionSubmissionRound,
    TransactionSubmissionAbciApp,
)


abci_app_transition_mapping: AbciAppTransitionMapping = {
    FinishedRegistrationRound: RandomnessSafeRound,
    FinishedSafeRound: StrategyEvaluationRound,
    FinishedRegistrationFFWRound: StrategyEvaluationRound,
    FinishedEnterPoolTransactionHashRound: RandomnessTransactionSubmissionRound,
    FinishedExitPoolTransactionHashRound: RandomnessTransactionSubmissionRound,
    FinishedSwapBackTransactionHashRound: RandomnessTransactionSubmissionRound,
    FinishedTransactionSubmissionRound: ResetAndPauseRound,
    FinishedResetAndPauseRound: StrategyEvaluationRound,
    FinishedResetAndPauseErrorRound: RegistrationRound,
    FailedRound: RegistrationRound,
}

LiquidityProvisionAbciApp = chain(
    (
        AgentRegistrationAbciApp,
        SafeDeploymentAbciApp,
        LiquidityRebalancingAbciApp,
        TransactionSubmissionAbciApp,
        ResetPauseABCIApp,
    ),
    abci_app_transition_mapping,
)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the transaction payloads for the oracle deployment app."""

from enum import Enum
from typing import Any, Dict, Optional

from packages.valory.skills.abstract_round_abci.base import BaseTxPayload


class TransactionType(Enum):
    """Enumeration of transaction types."""

    DEPLOY_ORACLE = "deploy_oracle"
    VALIDATE = "validate_oracle"
    RANDOMNESS = "randomness_oracle"
    SELECT_KEEPER = "select_keeper_oracle"

    def __str__(self) -> str:
        """Get the string value of the transaction type."""
        return self.value


class RandomnessPayload(BaseTxPayload):
    """Represent a transaction payload of type 'randomness'."""

    transaction_type = TransactionType.RANDOMNESS

    def __init__(
        self, sender: str, round_id: int, randomness: str, **kwargs: Any
    ) -> None:
        """Initialize an 'select_keeper' transaction payload.

        :param sender: the sender (Ethereum) address
        :param round_id: the round id
        :param randomness: the randomness
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._round_id = round_id
        self._randomness = randomness

    @property
    def round_id(self) -> int:
        """Get the round id."""
        return self._round_id  # pragma: nocover

    @property
    def randomness(self) -> str:
        """Get the randomness."""
        return self._randomness

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(round_id=self._round_id, randomness=self._randomness)


class SelectKeeperPayload(BaseTxPayload):
    """Represent a transaction payload of type 'select_keeper'."""

    transaction_type = TransactionType.SELECT_KEEPER

    def __init__(self, sender: str, keeper: str, **kwargs: Any) -> None:
        """Initialize an 'select_keeper' transaction payload.

        :param sender: the sender (Ethereum) address
        :param keeper: the keeper selection
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._keeper = keeper

    @property
    def keeper(self) -> str:
        """Get the keeper."""
        return self._keeper

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(keeper=self.keeper)


class ValidateOraclePayload(BaseTxPayload):
    """Represent a transaction payload of type 'validate'."""

    transaction_type = TransactionType.VALIDATE

    def __init__(self, sender: str, vote: Optional[bool] = None, **kwargs: Any) -> None:
        """Initialize an 'validate' transaction payload.

        :param sender: the sender (Ethereum) address
        :param vote: the vote
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._vote = vote

    @property
    def vote(self) -> Optional[bool]:
        """Get the vote."""
        return self._vote

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(vote=self.vote) if self.vote is not None else {}


class DeployOraclePayload(BaseTxPayload):
    """Represent a transaction payload of type 'deploy_oracle'."""

    transaction_type = TransactionType.DEPLOY_ORACLE

    def __init__(
        self, sender: str, oracle_contract_address: str, **kwargs: Any
    ) -> None:
        """Initialize a 'deploy_safe' transaction payload.

        :param sender: the sender (Ethereum) address
        :param oracle_contract_address: the Safe contract address
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._oracle_contract_address = oracle_contract_address

    @property
    def oracle_contract_address(self) -> str:
        """Get the Oracle contract address."""
        return self._oracle_contract_address

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(oracle_contract_address=self.oracle_contract_address)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the models.py module of the skill."""
from unittest import mock
from unittest.mock import MagicMock

import pytest

from packages.valory.skills.test_abci.models import SharedState
from packages.valory.skills.test_abci.rounds import Event, TestAbciApp


@pytest.fixture
def shared_state() -> SharedState:
    """Initialize a test shared state."""
    return SharedState(name="", skill_context=mock.MagicMock())


class DummyContext:
    """Dummy Context class for shared state."""

    class params:
        """Dummy param variable."""

        round_timeout_seconds: float = 1.0


class TestSharedState:
    """Test SharedState(Model) class."""

    def test_initialization(
        self,
    ) -> None:
        """Test initialization."""
        SharedState(name="", skill_context=DummyContext())

    def test_setup(
        self,
        shared_state: SharedState,
    ) -> None:
        """Test setup."""
        shared_state.context.params.setup_params = {"test": []}
        shared_state.context.params.consensus_params = MagicMock()
        shared_state.setup()
        assert (
            TestAbciApp.event_to_timeout[Event.ROUND_TIMEOUT]
            == shared_state.context.params.round_timeout_seconds
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Integration tests for the valory/register_reset skill."""
import pytest
from aea.configurations.data_types import PublicId

from tests.test_agents.base import (
    BaseTestEnd2EndAgentCatchup,
    BaseTestEnd2EndNormalExecution,
)


# round check log messages of the happy path
EXPECTED_ROUND_LOG_COUNT = {
    "registration_startup": 1,
    "registration": 3,
    "reset_and_pause": 4,
}


@pytest.mark.parametrize("nb_nodes", (4,))
class TestTendermintStartup(BaseTestEnd2EndNormalExecution):
    """Test the ABCI register-reset skill with 4 agents starting up."""

    agent_package = "valory/register_reset:0.1.0"
    skill_package = "valory/register_reset_abci:0.1.0"
    round_check_strings_to_n_periods = {
        "registration_startup": 1,
        "reset_and_pause": 1,
    }
    wait_to_finish = 60


@pytest.mark.parametrize("nb_nodes", (4,))
class TestTendermintReset(BaseTestEnd2EndNormalExecution):
    """Test the ABCI register-reset skill with 4 agents when resetting Tendermint."""

    agent_package = "valory/register_reset:0.1.0"
    skill_package = "valory/register_reset_abci:0.1.0"
    round_check_strings_to_n_periods = EXPECTED_ROUND_LOG_COUNT
    wait_to_finish = 200
    __reset_tendermint_every = 1
    __args_prefix = f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.models.params.args"
    # reset every `__reset_tendermint_every` rounds
    extra_configs = [
        {
            "dotted_path": f"{__args_prefix}.reset_tendermint_after",
            "value": __reset_tendermint_every,
        },
    ]


@pytest.mark.parametrize("nb_nodes", (4,))
class TestTendermintResetInterrupt(BaseTestEnd2EndAgentCatchup):
    """Test the ABCI register-reset skill with 4 agents when an agent gets interrupted on Tendermint reset."""

    agent_package = "valory/register_reset:0.1.0"
    skill_package = "valory/register_reset_abci:0.1.0"
    cli_log_options = ["-v", "INFO"]
    wait_before_stop = 60
    wait_to_finish = 200
    restart_after = 1
    __reset_tendermint_every = 1
    stop_string = f"Entered in the 'reset_and_pause' round for period {__reset_tendermint_every - 1}"
    round_check_strings_to_n_periods = EXPECTED_ROUND_LOG_COUNT

    __args_prefix = f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.models.params.args"
    # reset every `__reset_tendermint_every` rounds
    extra_configs = [
        {
            "dotted_path": f"{__args_prefix}.reset_tendermint_after",
            "value": __reset_tendermint_every,
        },
    ]


class TestTendermintResetInterruptNoRejoin(TestTendermintResetInterrupt):
    """
    Test a Tendermint reset case for the ABCI register-reset skill.

    Test the ABCI register-reset skill with 4 agents when an agent gets temporarily interrupted
    on Tendermint reset and never rejoins.
    """

    wait_to_finish = 200
    # set the restart to a value so that the agent never rejoins, in order to test the impact to the rest of the agents
    restart_after = wait_to_finish
    # check if we manage to reset with Tendermint with the rest of the agents; 3rd agent will not rejoin in this test
    exclude_from_checks = [3]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the transaction payloads for common apps."""
from enum import Enum
from typing import Any, Dict, Optional, Union

from packages.valory.skills.abstract_round_abci.base import BaseTxPayload


class TransactionType(Enum):
    """Enumeration of transaction types."""

    VALIDATE = "validate_transaction"
    SIGNATURE = "signature_transaction"
    FINALIZATION = "finalization_transaction"
    RESET = "reset_transaction"
    RANDOMNESS = "randomness_transaction"
    SELECT_KEEPER = "select_keeper_transaction"
    CHECK = "check"
    SYNCHRONIZE = "synchronize"

    def __str__(self) -> str:
        """Get the string value of the transaction type."""
        return self.value


class RandomnessPayload(BaseTxPayload):
    """Represent a transaction payload of type 'randomness'."""

    transaction_type = TransactionType.RANDOMNESS

    def __init__(
        self, sender: str, round_id: int, randomness: str, **kwargs: Any
    ) -> None:
        """Initialize an 'select_keeper' transaction payload.

        :param sender: the sender (Ethereum) address
        :param round_id: the round id
        :param randomness: the randomness
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._round_id = round_id
        self._randomness = randomness

    @property
    def round_id(self) -> int:
        """Get the round id."""
        return self._round_id  # pragma: nocover

    @property
    def randomness(self) -> str:
        """Get the randomness."""
        return self._randomness  # pragma: nocover

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(round_id=self._round_id, randomness=self._randomness)


class SelectKeeperPayload(BaseTxPayload):
    """Represent a transaction payload of type 'select_keeper'."""

    transaction_type = TransactionType.SELECT_KEEPER

    def __init__(self, sender: str, keepers: str, **kwargs: Any) -> None:
        """Initialize a 'select_keeper' transaction payload.

        :param sender: the sender (Ethereum) address
        :param keepers: the updated keepers
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._keepers = keepers

    @property
    def keepers(self) -> str:
        """Get the keeper."""
        return self._keepers

    @property
    def data(self) -> Dict[str, str]:
        """Get the data."""
        return dict(keepers=self.keepers)


class ValidatePayload(BaseTxPayload):
    """Represent a transaction payload of type 'validate'."""

    transaction_type = TransactionType.VALIDATE

    def __init__(self, sender: str, vote: Optional[bool] = None, **kwargs: Any) -> None:
        """Initialize an 'validate' transaction payload.

        :param sender: the sender (Ethereum) address
        :param vote: the vote
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._vote = vote

    @property
    def vote(self) -> Optional[bool]:
        """Get the vote."""
        return self._vote

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(vote=self.vote) if self.vote is not None else {}


class CheckTransactionHistoryPayload(BaseTxPayload):
    """Represent a transaction payload of type 'check'."""

    transaction_type = TransactionType.CHECK

    def __init__(self, sender: str, verified_res: str, **kwargs: Any) -> None:
        """Initialize an 'check' transaction payload.

        :param sender: the sender (Ethereum) address
        :param verified_res: the verification result
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._verified_res = verified_res

    @property
    def verified_res(self) -> str:
        """Get the verified result."""
        return self._verified_res

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(verified_res=self.verified_res)


class SynchronizeLateMessagesPayload(BaseTxPayload):
    """Represent a transaction payload of type 'synchronize'."""

    transaction_type = TransactionType.SYNCHRONIZE

    def __init__(self, sender: str, tx_hashes: str = "", **kwargs: Any) -> None:
        """Initialize a 'synchronize' transaction payload.

        :param sender: the sender (Ethereum) address
        :param tx_hashes: the late-arriving tx hashes concatenated
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._tx_hashes = tx_hashes

    @property
    def tx_hashes(self) -> Optional[str]:
        """Get the late-arriving tx hashes."""
        return None if self._tx_hashes == "" else self._tx_hashes

    @property
    def data(self) -> Dict[str, Optional[str]]:
        """Get the data."""
        return dict(tx_hashes=self._tx_hashes)


class SignaturePayload(BaseTxPayload):
    """Represent a transaction payload of type 'signature'."""

    transaction_type = TransactionType.SIGNATURE

    def __init__(self, sender: str, signature: str, **kwargs: Any) -> None:
        """Initialize an 'signature' transaction payload.

        :param sender: the sender (Ethereum) address
        :param signature: the signature
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._signature = signature

    @property
    def signature(self) -> str:
        """Get the signature."""
        return self._signature

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(signature=self.signature)


class FinalizationTxPayload(BaseTxPayload):
    """Represent a transaction payload of type 'finalization'."""

    transaction_type = TransactionType.FINALIZATION

    def __init__(
        self,
        sender: str,
        tx_data: Optional[Dict[str, Union[str, int, bool]]] = None,
        **kwargs: Any
    ) -> None:
        """Initialize an 'finalization' transaction payload.

        :param sender: the sender (Ethereum) address
        :param tx_data: the transaction data
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._tx_data = tx_data

    @property
    def tx_data(self) -> Optional[Dict[str, Union[str, int, bool]]]:
        """Get the tx_data."""
        return self._tx_data

    @property
    def data(self) -> Dict[str, Dict[str, Union[str, int, bool]]]:
        """Get the data."""
        return dict(tx_data=self._tx_data) if self._tx_data is not None else {}


class ResetPayload(BaseTxPayload):
    """Represent a transaction payload of type 'reset'."""

    transaction_type = TransactionType.RESET

    def __init__(self, sender: str, period_count: int, **kwargs: Any) -> None:
        """Initialize an 'reset' transaction payload.

        :param sender: the sender (Ethereum) address
        :param period_count: the period count id
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._period_count = period_count

    @property
    def period_count(self) -> int:
        """Get the period_count."""
        return self._period_count

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(period_count=self.period_count)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------


"""Tools for the APY skill."""

from typing import Iterator, Optional


def gen_unix_timestamps(synced_now: int, duration: int) -> Iterator[int]:
    """Generate the UNIX timestamps from `duration` months ago up to today.

    :param synced_now: the synced time across the agents.
    :param duration: the duration of the timestamps to be returned, in months (more precisely, in 30 days).
    :yields: the UNIX timestamps.
    """
    day_in_unix = 24 * 60 * 60

    duration_before = synced_now - (duration * 30 * day_in_unix)

    for day in range(duration_before, synced_now, day_in_unix):
        yield day


def filter_out_numbers(string: str) -> Optional[int]:
    """Filter out all the numbers from a string.

    :param string: the string to filter.
    :return: a filtered out integer.
    """
    numeric_filter = filter(str.isdigit, string)
    numeric_string = "".join(numeric_filter)

    if numeric_string == "":
        filtered_result = None
    else:
        filtered_result = int(numeric_string)
        # Get only the first 9 digits, because the seed cannot be > 2**32 -1
        filtered_result = int(str(filtered_result)[:9])

    return filtered_result



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the transaction payloads for the liquidity_rebalancing_abci skill."""
from abc import ABC
from enum import Enum
from typing import Any, Dict, Optional

from packages.valory.skills.abstract_round_abci.base import BaseTxPayload


class TransactionType(Enum):
    """Enumeration of transaction types."""

    STRATEGY_EVALUATION = "strategy_evaluation"
    WAIT = "wait"
    ALLOWANCE_CHECK = "allowance_check"
    TRANSACTION_HASH = "tx_hash"
    LP_RESULT = "lp_result"
    TX_HASH = "tx_hash"
    SLEEP = "sleep"

    def __str__(self) -> str:
        """Get the string value of the transaction type."""
        return self.value


class BaseLiquidityRebalancingPayload(BaseTxPayload, ABC):
    """Base class for the liquidity rebalancing skill."""

    def __hash__(self) -> int:  # pragma: nocover
        """Hash the payload."""
        return hash(tuple(sorted(self.data.items())))


class StrategyType(Enum):
    """Enumeration of strategy types."""

    WAIT = "wait"
    ENTER = "enter"
    EXIT = "exit"
    SWAP_BACK = "swap_back"

    def __str__(self) -> str:  # pragma: nocover
        """Get the string value of the strategy type."""
        return self.value


class StrategyEvaluationPayload(BaseLiquidityRebalancingPayload):
    """Represent a transaction payload of type 'strategy_evaluation'."""

    transaction_type = TransactionType.STRATEGY_EVALUATION

    def __init__(self, sender: str, strategy: str, **kwargs: Any) -> None:
        """Initialize a 'strategy_evaluation' transaction payload.

        :param sender: the sender (Ethereum) address
        :param strategy: the new strategy to follow
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._strategy = strategy

    @property
    def strategy(self) -> str:
        """Get the strategy."""
        return self._strategy

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(strategy=self.strategy)


class TransactionHashPayload(BaseTxPayload):
    """Represent a transaction payload of type 'tx_hash'."""

    transaction_type = TransactionType.TX_HASH

    def __init__(
        self, sender: str, tx_hash: Optional[str] = None, **kwargs: Any
    ) -> None:
        """Initialize an 'tx_hash' transaction payload.

        :param sender: the sender (Ethereum) address
        :param tx_hash: the tx_hash
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._tx_hash = tx_hash

    @property
    def tx_hash(self) -> Optional[str]:
        """Get the tx_hash."""
        return self._tx_hash

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(tx_hash=self.tx_hash) if self.tx_hash is not None else {}


class SleepPayload(BaseTxPayload):
    """Represent a transaction payload of type 'sleep'."""

    transaction_type = TransactionType.SLEEP

    @property
    def sleep(self) -> str:
        """Get the sleep dummy property."""
        return "sleep"

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(sleep=self.sleep)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the APY estimation ABCI application."""

from packages.valory.skills.abstract_round_abci.abci_app_chain import (
    AbciAppTransitionMapping,
    chain,
)
from packages.valory.skills.apy_estimation_abci.rounds import (
    APYEstimationAbciApp,
    CollectHistoryRound,
    FailedAPYRound,
)
from packages.valory.skills.registration_abci.rounds import (
    AgentRegistrationAbciApp,
    FinishedRegistrationFFWRound,
    FinishedRegistrationRound,
    RegistrationRound,
)


abci_app_transition_mapping: AbciAppTransitionMapping = {
    FinishedRegistrationRound: CollectHistoryRound,
    FinishedRegistrationFFWRound: CollectHistoryRound,
    FailedAPYRound: RegistrationRound,
}

APYEstimationAbciAppChained = chain(
    (
        AgentRegistrationAbciApp,
        APYEstimationAbciApp,
    ),
    abci_app_transition_mapping,
)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/transaction settlement skill's payload tools."""
import sys

import atheris  # type: ignore
import pytest

from packages.valory.contracts.gnosis_safe.contract import SafeOperation
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    NULL_ADDRESS,
    PayloadDeserializationError,
    VerificationStatus,
    hash_payload_to_hex,
    skill_input_hex_to_payload,
    tx_hist_hex_to_payload,
    tx_hist_payload_to_hex,
)


class TestTxHistPayloadEncodingDecoding:
    """Tests for the transaction history's payload encoding - decoding."""

    @staticmethod
    @pytest.mark.parametrize(
        "verification_status, tx_hash",
        (
            (
                VerificationStatus.VERIFIED,
                "0xb0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
            ),
            (VerificationStatus.ERROR, None),
        ),
    )
    def test_tx_hist_payload_to_hex_and_back(
        verification_status: VerificationStatus, tx_hash: str
    ) -> None:
        """Test `tx_hist_payload_to_hex` and `tx_hist_hex_to_payload` functions."""
        intermediate = tx_hist_payload_to_hex(verification_status, tx_hash)
        verification_status_, tx_hash_ = tx_hist_hex_to_payload(intermediate)
        assert verification_status == verification_status_
        assert tx_hash == tx_hash_

    @staticmethod
    def test_invalid_tx_hash_during_serialization() -> None:
        """Test encoding when transaction hash is invalid."""
        with pytest.raises(ValueError):
            tx_hist_payload_to_hex(VerificationStatus.VERIFIED, "")

    @staticmethod
    @pytest.mark.parametrize(
        "payload",
        ("0000000000000000000000000000000000000000000000000000000000000007", ""),
    )
    def test_invalid_payloads_during_deserialization(payload: str) -> None:
        """Test decoding payload is invalid."""
        with pytest.raises(PayloadDeserializationError):
            tx_hist_hex_to_payload(payload)


def test_payload_to_hex_and_back() -> None:
    """Test `payload_to_hex` function."""
    tx_params = dict(
        safe_tx_hash="b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
        ether_value=0,
        safe_tx_gas=40000000,
        to_address="0x77E9b2EF921253A171Fa0CB9ba80558648Ff7215",
        data=b"b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
        operation=SafeOperation.CALL.value,
        base_gas=0,
        safe_gas_price=0,
        gas_token=NULL_ADDRESS,
        refund_receiver=NULL_ADDRESS,
    )

    intermediate = hash_payload_to_hex(**tx_params)  # type: ignore
    assert tx_params == skill_input_hex_to_payload(intermediate)


@pytest.mark.skip
def test_fuzz_tx_hist_payload_to_hex() -> None:
    """Test fuzz tx_hist_payload_to_hex."""

    @atheris.instrument_func
    def fuzz_tx_hist_payload_to_hex(input_bytes: bytes) -> None:
        """Fuzz tx_hist_payload_to_hex."""
        fdp = atheris.FuzzedDataProvider(input_bytes)
        verification_int = fdp.ConsumeInt(4)
        if verification_int < 1 or verification_int > 6:
            return
        verification = VerificationStatus(verification_int)
        tx_hash = fdp.ConsumeString(12)
        if len(tx_hash) != 64:
            return
        tx_hist_payload_to_hex(verification, tx_hash)

    atheris.instrument_all()
    atheris.Setup(sys.argv, fuzz_tx_hist_payload_to_hex)
    atheris.Fuzz()


@pytest.mark.skip
def test_fuzz_tx_hist_hex_to_payload() -> None:
    """Test fuzz tx_hist_hex_to_payload."""

    @atheris.instrument_func
    def fuzz_tx_hist_hex_to_payload(input_bytes: bytes) -> None:
        """Fuzz tx_hist_hex_to_payload."""
        fdp = atheris.FuzzedDataProvider(input_bytes)
        try:
            payload = fdp.ConsumeString(12)
            tx_hist_hex_to_payload(payload)
        except PayloadDeserializationError:
            pass

    atheris.instrument_all()
    atheris.Setup(sys.argv, fuzz_tx_hist_hex_to_payload)
    atheris.Fuzz()


@pytest.mark.skip
def test_fuzz_hash_payload_to_hex() -> None:
    """Test fuzz hash_payload_to_hex."""

    @atheris.instrument_func
    def fuzz_hash_payload_to_hex(input_bytes: bytes) -> None:
        """Fuzz hash_payload_to_hex."""
        fdp = atheris.FuzzedDataProvider(input_bytes)

        safe_tx_hash = fdp.ConsumeString(64)
        ether_value = fdp.ConsumeInt(4)
        safe_tx_gas = fdp.ConsumeInt(4)
        to_address = fdp.ConsumeString(42)
        data = fdp.ConsumeBytes(12)
        operation = fdp.ConsumeInt(4)
        base_gas = fdp.ConsumeInt(4)
        safe_gas_price = fdp.ConsumeInt(4)

        if len(safe_tx_hash) != 64:
            return

        if len(to_address) != 42:
            return

        hash_payload_to_hex(
            safe_tx_hash,
            ether_value,
            safe_tx_gas,
            to_address,
            data,
            operation,
            base_gas,
            safe_gas_price,
        )

    atheris.instrument_all()
    atheris.Setup(sys.argv, fuzz_hash_payload_to_hex)
    atheris.Fuzz()


@pytest.mark.skip
def test_fuzz_skill_input_hex_to_payload() -> None:
    """Test fuzz skill_input_hex_to_payload."""

    @atheris.instrument_func
    def fuzz_skill_input_hex_to_payload(input_bytes: bytes) -> None:
        """Fuzz skill_input_hex_to_payload."""
        fdp = atheris.FuzzedDataProvider(input_bytes)
        payload = fdp.ConsumeString(500)
        if len(payload) < 500:
            return
        try:
            skill_input_hex_to_payload(payload)
        except TypeError:
            pass

    atheris.instrument_all()
    atheris.Setup(sys.argv, fuzz_skill_input_hex_to_payload)
    atheris.Fuzz()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/abci connection, tendermint_encoder module."""

from packages.valory.connections.abci.tendermint.abci.types_pb2 import (  # type: ignore
    ResponseListSnapshots,
)
from packages.valory.connections.abci.tendermint_encoder import (
    _TendermintProtocolEncoder,
)
from packages.valory.protocols.abci import AbciMessage
from packages.valory.protocols.abci.custom_types import Result, ResultType, SnapShots
from packages.valory.protocols.abci.custom_types import Snapshot as CustomSnapshot


class TestTendermintProtocolEncoder:
    """Test for the Tendermint protocol encoder."""

    def test_response_exception(self) -> None:
        """Test decoding of a response exception."""
        expected_error = "error"
        abci_message = AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_EXCEPTION,  # type: ignore
            error=expected_error,
        )
        message = _TendermintProtocolEncoder.response_exception(abci_message)
        assert message.exception.error == expected_error

    def test_response_echo(self) -> None:
        """Test decoding of a response echo."""
        expected_message = "message"
        abci_message = AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_ECHO,  # type: ignore
            message=expected_message,
        )
        message = _TendermintProtocolEncoder.response_echo(abci_message)
        assert message.echo.message == expected_message

    def test_response_set_option(self) -> None:
        """Test decoding of a response set-option."""
        expected_code = 0
        expected_log = "log"
        expected_info = "info"
        abci_message = AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_SET_OPTION,  # type: ignore
            code=expected_code,
            log=expected_log,
            info=expected_info,
        )
        message = _TendermintProtocolEncoder.response_set_option(abci_message)
        assert message.set_option.code == expected_code
        assert message.set_option.log == expected_log
        assert message.set_option.info == expected_info

    def test_response_list_snapshots(self) -> None:
        """Test decoding of a response list-snapshots."""
        snapshot = CustomSnapshot(0, 0, 0, b"", b"")
        snapshots = SnapShots([snapshot])

        # expected snapshots object
        list_snapshots = ResponseListSnapshots()
        snapshots_pb = [
            _TendermintProtocolEncoder._encode_snapshot(snapshot)
            for snapshot in snapshots.snapshots
        ]
        list_snapshots.snapshots.extend(snapshots_pb)

        abci_message = AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_LIST_SNAPSHOTS,  # type: ignore
            snapshots=snapshots,
        )
        message = _TendermintProtocolEncoder.response_list_snapshots(abci_message)
        assert message.list_snapshots == list_snapshots

    def test_response_offer_snapshot(self) -> None:
        """Test decoding of a response offer-snapshot."""
        expected_result = Result(ResultType.ACCEPT)
        abci_message = AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_OFFER_SNAPSHOT,  # type: ignore
            result=expected_result,
        )
        message = _TendermintProtocolEncoder.response_offer_snapshot(abci_message)
        assert message.offer_snapshot.result == expected_result.result_type.value

    def test_response_load_snapshot_chunk(self) -> None:
        """Test decoding of a response load-snapshot-chunk."""
        expected_chunk = b"chunk"
        abci_message = AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_LOAD_SNAPSHOT_CHUNK,  # type: ignore
            chunk=expected_chunk,
        )
        message = _TendermintProtocolEncoder.response_load_snapshot_chunk(abci_message)
        assert message.load_snapshot_chunk.chunk == expected_chunk

    def test_response_apply_snapshot_chunk(self) -> None:
        """Test decoding of a response apply-snapshot-chunk."""
        result = Result(ResultType.ACCEPT)
        abci_message = AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_APPLY_SNAPSHOT_CHUNK,  # type: ignore
            result=result,
            refetch_chunks=[],
            reject_senders=[],
        )
        message = _TendermintProtocolEncoder.response_apply_snapshot_chunk(abci_message)
        assert message.apply_snapshot_chunk.result == result.result_type.value



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests package for the 'valory/abci' protocol."""
import datetime
from abc import abstractmethod
from typing import Callable, Type
from unittest import mock

import pytest
from aea.common import Address
from aea.exceptions import AEAEnforceError
from aea.mail.base import Envelope
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.protocols.dialogue.base import DialogueLabel

import packages
from packages.valory.protocols.abci import AbciMessage
from packages.valory.protocols.abci.custom_types import (
    BlockID,
    BlockParams,
    CheckTxType,
    CheckTxTypeEnum,
    ConsensusParams,
    ConsensusVersion,
    Duration,
    Event,
    EventAttribute,
    Events,
    Evidence,
    EvidenceParams,
    EvidenceType,
    Evidences,
    Header,
    LastCommitInfo,
    PartSetHeader,
    ProofOp,
    ProofOps,
    PublicKey,
    Result,
    ResultType,
    SnapShots,
    Snapshot,
    Timestamp,
    Validator,
    ValidatorParams,
    ValidatorUpdate,
    ValidatorUpdates,
    VersionParams,
    VoteInfo,
)
from packages.valory.protocols.abci.dialogues import AbciDialogue, AbciDialogues
from packages.valory.protocols.abci.message import (
    _default_logger as abci_message_logger,
)


class BaseTestMessageConstruction:
    """Base class to test message construction for the ABCI protocol."""

    @abstractmethod
    def build_message(self) -> AbciMessage:
        """Build the message to be used for testing."""

    def test_run(self) -> None:
        """Run the test."""
        msg = self.build_message()
        msg.to = "receiver"
        envelope = Envelope(to=msg.to, sender="sender", message=msg)
        envelope_bytes = envelope.encode()

        actual_envelope = Envelope.decode(envelope_bytes)
        expected_envelope = envelope

        assert expected_envelope.to == actual_envelope.to
        assert expected_envelope.sender == actual_envelope.sender
        assert (
            expected_envelope.protocol_specification_id
            == actual_envelope.protocol_specification_id
        )
        assert expected_envelope.message != actual_envelope.message

        actual_msg = AbciMessage.serializer.decode(actual_envelope.message_bytes)
        actual_msg.to = actual_envelope.to
        actual_msg.sender = actual_envelope.sender
        expected_msg = msg
        assert expected_msg == actual_msg

    @classmethod
    def _make_consensus_params(cls) -> ConsensusParams:
        """Build a ConsensuParams object."""
        return ConsensusParams(
            BlockParams(0, 0),
            EvidenceParams(0, Duration(0, 0), 0),
            ValidatorParams(["pub_key"]),
            VersionParams(0),
        )

    @classmethod
    def _make_snapshot(cls) -> Snapshot:
        """Build a Snapshot object."""
        return Snapshot(0, 0, 0, b"hash", b"metadata")


class TestRequestEcho(BaseTestMessageConstruction):
    """Test ABCI request abci."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_ECHO,  # type: ignore
            message="hello",
        )


class TestResponseEcho(BaseTestMessageConstruction):
    """Test ABCI response abci."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_ECHO,  # type: ignore
            message="hello",
        )


class TestRequestFlush(BaseTestMessageConstruction):
    """Test ABCI request flush."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_FLUSH,  # type: ignore
        )


class TestResponseFlush(BaseTestMessageConstruction):
    """Test ABCI response flush."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_FLUSH,  # type: ignore
        )


class TestRequestInfo(BaseTestMessageConstruction):
    """Test ABCI request info."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_INFO,  # type: ignore
            version="0.1.0",
            block_version=1,
            p2p_version=1,
        )


class TestResponseInfo(BaseTestMessageConstruction):
    """Test ABCI response info."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_INFO,  # type: ignore
            info_data="info",
            version="0.1.0",
            app_version=1,
            last_block_height=1,
            last_block_app_hash=b"bytes",
        )


class TestRequestInitChain(BaseTestMessageConstruction):
    """Test ABCI request init chain."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        consensus_params = super()._make_consensus_params()

        validators = ValidatorUpdates(
            [
                ValidatorUpdate(
                    PublicKey(b"pub_key_bytes", PublicKey.PublicKeyType.ed25519), 1
                ),
                ValidatorUpdate(
                    PublicKey(b"pub_key_bytes", PublicKey.PublicKeyType.secp256k1), 2
                ),
            ]
        )

        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_INIT_CHAIN,  # type: ignore
            time=Timestamp(0, 0),
            chain_id="1",
            consensus_params=consensus_params,
            validators=validators,
            app_state_bytes=b"bytes",
            initial_height=0,
        )


class TestResponseInitChain(BaseTestMessageConstruction):
    """Test ABCI response init chain."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        consensus_params = super()._make_consensus_params()

        validators = ValidatorUpdates(
            [
                ValidatorUpdate(
                    PublicKey(b"pub_key_bytes", PublicKey.PublicKeyType.ed25519), 1
                ),
                ValidatorUpdate(
                    PublicKey(b"pub_key_bytes", PublicKey.PublicKeyType.secp256k1), 2
                ),
            ]
        )

        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_INIT_CHAIN,  # type: ignore
            consensus_params=consensus_params,
            validators=validators,
            app_hash=b"app_hash",
        )


class TestRequestQuery(BaseTestMessageConstruction):
    """Test ABCI response query."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_QUERY,  # type: ignore
            query_data=b"bytes",
            path="",
            height=0,
            prove=False,
        )


class TestResponseQuery(BaseTestMessageConstruction):
    """Test ABCI response query."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_QUERY,  # type: ignore
            code=0,
            log="log",
            info="info",
            index=0,
            key=b"key",
            value=b"value",
            proof_ops=ProofOps(
                [
                    ProofOp("type", b"key", b"data"),
                    ProofOp("type", b"key", b"data"),
                ]
            ),
            height=0,
            codespace="",
        )


class TestRequestBeginBlock(BaseTestMessageConstruction):
    """Test ABCI request begin block."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        header = Header(
            ConsensusVersion(0, 0),
            "chain_id",
            0,
            Timestamp(0, 0),
            BlockID(b"hash", PartSetHeader(0, b"hash")),
            b"last_commit_hash",
            b"data_hash",
            b"validators_hash",
            b"next_validators_hash",
            b"consensus_hash",
            b"app_hash",
            b"last_results_hash",
            b"evidence_hash",
            b"proposer_address",
        )

        assert header.timestamp == datetime.datetime.fromtimestamp(0)

        validator = Validator(b"address", 0)
        last_commit_info = LastCommitInfo(
            0,
            [
                VoteInfo(validator, True),
                VoteInfo(validator, False),
            ],
        )

        evidences = Evidences(
            [
                Evidence(EvidenceType.UNKNOWN, validator, 0, Timestamp(0, 0), 0),
                Evidence(EvidenceType.DUPLICATE_VOTE, validator, 0, Timestamp(0, 0), 0),
            ]
        )

        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_BEGIN_BLOCK,  # type: ignore
            hash=b"hash",
            header=header,
            last_commit_info=last_commit_info,
            byzantine_validators=evidences,
        )


class TestResponseBeginBlock(BaseTestMessageConstruction):
    """Test ABCI response begin block."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        event = Event("type", [EventAttribute(b"key", b"value", True)])
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_BEGIN_BLOCK,  # type: ignore
            events=Events([event, event]),
        )


class TestRequestCheckTx(BaseTestMessageConstruction):
    """Test ABCI request check tx."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        tx = b"bytes"
        type_ = CheckTxType(CheckTxTypeEnum.NEW)
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_CHECK_TX,  # type: ignore
            tx=tx,
            type=type_,
        )


class TestResponseCheckTx(BaseTestMessageConstruction):
    """Test ABCI response check tx."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        attribute = EventAttribute(b"key", b"value", True)
        event = Event("type", attributes=[attribute, attribute])
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_CHECK_TX,  # type: ignore
            code=0,
            data=b"data",
            log="log",
            info="info",
            gas_wanted=0,
            gas_used=0,
            events=Events([event, event]),
            codespace="codespace",
        )


class TestRequestDeliverTx(BaseTestMessageConstruction):
    """Test ABCI request deliver tx."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_DELIVER_TX,  # type: ignore
            tx=b"tx",
        )


class TestResponseDeliverTx(BaseTestMessageConstruction):
    """Test ABCI response deliver tx."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        attribute = EventAttribute(b"key", b"value", True)
        event = Event("type", attributes=[attribute, attribute])
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_DELIVER_TX,  # type: ignore
            code=0,
            data=b"data",
            log="log",
            info="info",
            gas_wanted=0,
            gas_used=0,
            events=Events([event, event]),
            codespace="codespace",
        )


class TestRequestEndBlock(BaseTestMessageConstruction):
    """Test ABCI request end block."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_END_BLOCK,  # type: ignore
            height=0,
        )


class TestRequestSetOption(BaseTestMessageConstruction):
    """Test ABCI request end block."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_SET_OPTION,  # type: ignore
            option_key="",
            option_value="",
        )


class TestResponseEndBlock(BaseTestMessageConstruction):
    """Test ABCI response end block."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        attribute = EventAttribute(b"key", b"value", True)
        event = Event("type", attributes=[attribute, attribute])
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_END_BLOCK,  # type: ignore
            validator_updates=ValidatorUpdates(
                [
                    ValidatorUpdate(
                        PublicKey(b"pub_key_bytes", PublicKey.PublicKeyType.ed25519), 1
                    ),
                    ValidatorUpdate(
                        PublicKey(b"pub_key_bytes", PublicKey.PublicKeyType.secp256k1),
                        2,
                    ),
                ]
            ),
            consensus_param_updates=super()._make_consensus_params(),
            events=Events([event, event]),
        )


class TestRequestCommit(BaseTestMessageConstruction):
    """Test ABCI request commit."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_COMMIT,  # type: ignore
        )


class TestResponseCommit(BaseTestMessageConstruction):
    """Test ABCI response commit."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_COMMIT,  # type: ignore
            data=b"bytes",
            retain_height=0,
        )


class TestRequestListSnapshots(BaseTestMessageConstruction):
    """Test ABCI request list snapshots."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_LIST_SNAPSHOTS,  # type: ignore
        )


class TestResponseListSnapshots(BaseTestMessageConstruction):
    """Test ABCI response list snapshots."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        snapshots = SnapShots(
            [
                super()._make_snapshot(),
                super()._make_snapshot(),
            ]
        )
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_LIST_SNAPSHOTS,  # type: ignore
            snapshots=snapshots,
        )


class TestRequestOfferSnapshot(BaseTestMessageConstruction):
    """Test ABCI request offer snapshot."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_OFFER_SNAPSHOT,  # type: ignore
            snapshot=super()._make_snapshot(),
            app_hash=b"app_hash",
        )


class TestResponseOfferSnapshot(BaseTestMessageConstruction):
    """Test ABCI response offer snapshot."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_OFFER_SNAPSHOT,  # type: ignore
            result=Result(ResultType.UNKNOWN),
        )


class TestRequestLoadSnapshotChunk(BaseTestMessageConstruction):
    """Test ABCI request load snapshot chunk."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_LOAD_SNAPSHOT_CHUNK,  # type: ignore
            height=0,
            format=0,
            chunk_index=0,
        )


class TestResponseLoadSnapshotChunk(BaseTestMessageConstruction):
    """Test ABCI response load snapshot chunk."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_LOAD_SNAPSHOT_CHUNK,  # type: ignore
            chunk=b"chunk",
        )


class TestRequestApplySnapshotChunk(BaseTestMessageConstruction):
    """Test ABCI request load snapshot chunk."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.REQUEST_APPLY_SNAPSHOT_CHUNK,  # type: ignore
            index=0,
            chunk=b"chunk",
            chunk_sender="sender",
        )


class TestResponseApplySnapshotChunk(BaseTestMessageConstruction):
    """Test ABCI response apply snapshot chunk."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_APPLY_SNAPSHOT_CHUNK,  # type: ignore
            result=Result(ResultType.REJECT),
            refetch_chunks=(0, 1, 2),
            reject_senders=("sender_1", "sender_2"),
        )


class TestDummy(BaseTestMessageConstruction):
    """Test ABCI request abci."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.DUMMY,  # type: ignore
            dummy_consensus_params=self._make_consensus_params(),
        )


class TestResponseException(BaseTestMessageConstruction):
    """Test ABCI request abci."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_EXCEPTION, error=""  # type: ignore
        )


class TestResponseSetOption(BaseTestMessageConstruction):
    """Test ABCI request end block."""

    def build_message(self) -> AbciMessage:
        """Build the message."""
        return AbciMessage(
            performative=AbciMessage.Performative.RESPONSE_SET_OPTION,  # type: ignore
            code=1,
            log="",
            info="",
        )


@mock.patch.object(
    packages.valory.protocols.abci.message,
    "enforce",
    side_effect=AEAEnforceError("some error"),
)
def test_incorrect_message(mocked_enforce: Callable) -> None:
    """Test that we raise an exception when the message is incorrect."""
    with mock.patch.object(abci_message_logger, "error") as mock_logger:
        AbciMessage(
            message_id=1,
            dialogue_reference=(str(0), ""),
            target=0,
            performative=AbciMessage.Performative.DUMMY,  # type: ignore
        )

        mock_logger.assert_any_call("some error")


def test_performative_string_value() -> None:
    """Test the string valoe of performatives."""

    assert str(AbciMessage.Performative.DUMMY) == "dummy", "The str value must be dummy"
    assert (
        str(AbciMessage.Performative.REQUEST_APPLY_SNAPSHOT_CHUNK)
        == "request_apply_snapshot_chunk"
    ), "The str value must be request_apply_snapshot_chunk"
    assert (
        str(AbciMessage.Performative.REQUEST_BEGIN_BLOCK) == "request_begin_block"
    ), "The str value must be request_begin_block"
    assert (
        str(AbciMessage.Performative.REQUEST_CHECK_TX) == "request_check_tx"
    ), "The str value must be request_check_tx"
    assert (
        str(AbciMessage.Performative.REQUEST_COMMIT) == "request_commit"
    ), "The str value must be request_commit"
    assert (
        str(AbciMessage.Performative.REQUEST_DELIVER_TX) == "request_deliver_tx"
    ), "The str value must be request_deliver_tx"
    assert (
        str(AbciMessage.Performative.REQUEST_ECHO) == "request_echo"
    ), "The str value must be request_echo"
    assert (
        str(AbciMessage.Performative.REQUEST_END_BLOCK) == "request_end_block"
    ), "The str value must be request_end_block"
    assert (
        str(AbciMessage.Performative.REQUEST_FLUSH) == "request_flush"
    ), "The str value must be request_flush"
    assert (
        str(AbciMessage.Performative.REQUEST_INFO) == "request_info"
    ), "The str value must be request_info"
    assert (
        str(AbciMessage.Performative.REQUEST_INIT_CHAIN) == "request_init_chain"
    ), "The str value must be request_init_chain"
    assert (
        str(AbciMessage.Performative.REQUEST_LIST_SNAPSHOTS) == "request_list_snapshots"
    ), "The str value must be request_list_snapshots"
    assert (
        str(AbciMessage.Performative.REQUEST_LOAD_SNAPSHOT_CHUNK)
        == "request_load_snapshot_chunk"
    ), "The str value must be request_load_snapshot_chunk"
    assert (
        str(AbciMessage.Performative.REQUEST_OFFER_SNAPSHOT) == "request_offer_snapshot"
    ), "The str value must be request_offer_snapshot"
    assert (
        str(AbciMessage.Performative.REQUEST_QUERY) == "request_query"
    ), "The str value must be request_query"
    assert (
        str(AbciMessage.Performative.RESPONSE_APPLY_SNAPSHOT_CHUNK)
        == "response_apply_snapshot_chunk"
    ), "The str value must be response_apply_snapshot_chunk"
    assert (
        str(AbciMessage.Performative.RESPONSE_BEGIN_BLOCK) == "response_begin_block"
    ), "The str value must be response_begin_block"
    assert (
        str(AbciMessage.Performative.RESPONSE_CHECK_TX) == "response_check_tx"
    ), "The str value must be response_check_tx"
    assert (
        str(AbciMessage.Performative.RESPONSE_COMMIT) == "response_commit"
    ), "The str value must be response_commit"
    assert (
        str(AbciMessage.Performative.RESPONSE_DELIVER_TX) == "response_deliver_tx"
    ), "The str value must be response_deliver_tx"
    assert (
        str(AbciMessage.Performative.RESPONSE_ECHO) == "response_echo"
    ), "The str value must be response_echo"
    assert (
        str(AbciMessage.Performative.RESPONSE_END_BLOCK) == "response_end_block"
    ), "The str value must be response_end_block"
    assert (
        str(AbciMessage.Performative.RESPONSE_EXCEPTION) == "response_exception"
    ), "The str value must be response_exception"
    assert (
        str(AbciMessage.Performative.RESPONSE_FLUSH) == "response_flush"
    ), "The str value must be response_flush"
    assert (
        str(AbciMessage.Performative.RESPONSE_INFO) == "response_info"
    ), "The str value must be response_info"
    assert (
        str(AbciMessage.Performative.RESPONSE_INIT_CHAIN) == "response_init_chain"
    ), "The str value must be response_init_chain"
    assert (
        str(AbciMessage.Performative.RESPONSE_LIST_SNAPSHOTS)
        == "response_list_snapshots"
    ), "The str value must be response_list_snapshots"
    assert (
        str(AbciMessage.Performative.RESPONSE_LOAD_SNAPSHOT_CHUNK)
        == "response_load_snapshot_chunk"
    ), "The str value must be response_load_snapshot_chunk"
    assert (
        str(AbciMessage.Performative.RESPONSE_OFFER_SNAPSHOT)
        == "response_offer_snapshot"
    ), "The str value must be response_offer_snapshot"
    assert (
        str(AbciMessage.Performative.RESPONSE_QUERY) == "response_query"
    ), "The str value must be response_query"
    assert (
        str(AbciMessage.Performative.RESPONSE_SET_OPTION) == "response_set_option"
    ), "The str value must be response_set_option"
    assert (
        str(AbciMessage.Performative.REQUEST_SET_OPTION) == "request_set_option"
    ), "The str value must be request_set_option"


def test_encoding_unknown_performative() -> None:
    """Test that we raise an exception when the performative is unknown during encoding."""
    msg = AbciMessage(
        performative=AbciMessage.Performative.REQUEST_ECHO, message="Hello"  # type: ignore
    )

    with pytest.raises(ValueError, match="Performative not valid:"):
        with mock.patch.object(AbciMessage.Performative, "__eq__", return_value=False):
            AbciMessage.serializer.encode(msg)


def test_decoding_unknown_performative() -> None:
    """Test that we raise an exception when the performative is unknown during encoding."""
    msg = AbciMessage(
        performative=AbciMessage.Performative.REQUEST_ECHO, message="Hello"  # type: ignore
    )

    encoded_msg = AbciMessage.serializer.encode(msg)
    with pytest.raises(ValueError, match="Performative not valid:"):
        with mock.patch.object(AbciMessage.Performative, "__eq__", return_value=False):
            AbciMessage.serializer.decode(encoded_msg)


class AgentDialogue(AbciDialogue):
    """The dialogue class maintains state of a dialogue and manages it."""

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        self_address: Address,
        role: BaseDialogue.Role,
        message_class: Type[AbciMessage],
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for
        :param message_class: the message class
        """
        AbciDialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            self_address=self_address,
            role=role,
            message_class=message_class,
        )


class AgentDialogues(AbciDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, self_address: Address) -> None:
        """
        Initialize dialogues.

        :param self_address: the address of the entity for whom this dialogue is maintained
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return AbciDialogue.Role.CLIENT

        AbciDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            dialogue_class=AgentDialogue,
        )


class ServerDialogue(AbciDialogue):
    """The dialogue class maintains state of a dialogue and manages it."""

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        self_address: Address,
        role: BaseDialogue.Role,
        message_class: Type[AbciMessage],
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for
        :param message_class: the message class
        """
        AbciDialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            self_address=self_address,
            role=role,
            message_class=message_class,
        )


class ServerDialogues(AbciDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, self_address: Address) -> None:
        """
        Initialize dialogues.

        :param self_address: the address of the entity for whom this dialogue is maintained
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return AbciDialogue.Role.SERVER

        AbciDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            dialogue_class=ServerDialogue,
        )


class TestDialogues:
    """Tests abci dialogues."""

    agent_addr: str
    server_addr: str
    agent_dialogues: AgentDialogues
    server_dialogues: ServerDialogues

    @classmethod
    def setup_class(cls) -> None:
        """Set up the test."""
        cls.agent_addr = "agent address"
        cls.server_addr = "server address"
        cls.agent_dialogues = AgentDialogues(cls.agent_addr)
        cls.server_dialogues = ServerDialogues(cls.server_addr)

    def test_create_self_initiated(self) -> None:
        """Test the self initialisation of a dialogue."""
        result = self.agent_dialogues._create_self_initiated(
            dialogue_opponent_addr=self.server_addr,
            dialogue_reference=(str(0), ""),
            role=AbciDialogue.Role.CLIENT,
        )
        assert isinstance(result, AbciDialogue)
        assert result.role == AbciDialogue.Role.CLIENT, "The role must be client."

    def test_create_opponent_initiated(self) -> None:
        """Test the opponent initialisation of a dialogue."""
        result = self.agent_dialogues._create_opponent_initiated(
            dialogue_opponent_addr=self.server_addr,
            dialogue_reference=(str(0), ""),
            role=AbciDialogue.Role.CLIENT,
        )
        assert isinstance(result, AbciDialogue)
        assert result.role == AbciDialogue.Role.CLIENT, "The role must be client."



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the transaction payloads for the test_abci app."""
from abc import ABC
from enum import Enum

from packages.valory.skills.abstract_round_abci.base import BaseTxPayload


class TransactionType(Enum):
    """Enumeration of transaction types."""

    DUMMY = "dummy"

    def __str__(self) -> str:
        """Get the string value of the transaction type."""
        return self.value


class DummyPayload(BaseTxPayload, ABC):
    """Represent a transaction payload of type 'registration'."""

    transaction_type = TransactionType.DUMMY



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Integration tests for various transaction settlement skill's failure modes."""
from collections import deque
from math import ceil
from pathlib import Path
from typing import Any, Callable, Deque, Dict, Optional, cast
from unittest import mock

import pytest
from aea_ledger_ethereum import EthereumApi
from web3.types import RPCEndpoint, Wei

from packages.open_aea.protocols.signing import SigningMessage
from packages.open_aea.protocols.signing.custom_types import (
    RawTransaction,
    SignedMessage,
    SignedTransaction,
)
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.protocols.ledger_api import LedgerApiMessage
from packages.valory.protocols.ledger_api.custom_types import (
    State,
    TransactionDigest,
    TransactionReceipt,
)
from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.oracle_abci.behaviours import (
    OracleAbciAppConsensusBehaviour,
)
from packages.valory.skills.oracle_deployment_abci.behaviours import (
    DeployOracleBehaviour,
)
from packages.valory.skills.price_estimation_abci.behaviours import (
    TransactionHashBehaviour,
)
from packages.valory.skills.price_estimation_abci.rounds import (
    SynchronizedData as PriceEstimationSynchronizedSata,
)
from packages.valory.skills.transaction_settlement_abci.behaviours import (
    CheckLateTxHashesBehaviour,
    FinalizeBehaviour,
    SelectKeeperTransactionSubmissionBehaviourA,
    SelectKeeperTransactionSubmissionBehaviourB,
    SynchronizeLateMessagesBehaviour,
    TransactionSettlementBaseBehaviour,
)
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    VerificationStatus,
    hash_payload_to_hex,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    SynchronizedData as TxSettlementSynchronizedSata,
)

from tests.conftest import ROOT_DIR
from tests.test_skills.base import FSMBehaviourBaseCase
from tests.test_skills.integration import (
    ExpectedContentType,
    ExpectedTypesType,
    GnosisIntegrationBaseCase,
    HandlersType,
    IntegrationBaseCase,
)


SAFE_TX_GAS = 120000
ETHER_VALUE = 0
DUMMY_MAX_PRIORITY_FEE_PER_GAS = 3000000000
DUMMY_MAX_FEE_PER_GAS = 4000000000
DUMMY_REPRICING_MULTIPLIER = 1.1


class OracleBehaviourBaseCase(FSMBehaviourBaseCase):
    """Base case for testing the oracle."""

    path_to_skill = Path(ROOT_DIR, "packages", "valory", "skills", "oracle_abci")
    behaviour: OracleAbciAppConsensusBehaviour


class TransactionSettlementIntegrationBaseCase(
    OracleBehaviourBaseCase, GnosisIntegrationBaseCase
):
    """Base case for integration testing TransactionSettlement FSM Behaviour."""

    price_estimation_synchronized_data: PriceEstimationSynchronizedSata

    @classmethod
    def setup(cls, **kwargs: Any) -> None:
        """Setup."""
        super().setup()

        keeper_initial_retries = 1
        cls.tx_settlement_synchronized_data = TxSettlementSynchronizedSata(
            AbciAppDB(
                initial_data=AbciAppDB.data_to_lists(
                    dict(
                        safe_contract_address=cls.safe_contract_address,
                        participants=frozenset(list(cls.safe_owners.keys())),
                        keepers=keeper_initial_retries.to_bytes(32, "big").hex()
                        + cls.keeper_address,
                    )
                ),
            )
        )

        cls.price_estimation_synchronized_data = PriceEstimationSynchronizedSata(
            AbciAppDB(
                initial_data=AbciAppDB.data_to_lists(
                    dict(
                        safe_contract_address=cls.safe_contract_address,
                        participants=frozenset(list(cls.safe_owners.keys())),
                        most_voted_keeper_address=cls.keeper_address,
                        most_voted_estimate=1,
                    )
                ),
            )
        )

    def deploy_oracle(self) -> None:
        """Deploy the oracle."""
        cycles_enter = 4
        handlers_enter: HandlersType = [
            self.contract_handler,
            self.signing_handler,
            self.ledger_handler,
            self.ledger_handler,
        ]
        expected_content_enter: ExpectedContentType = [
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION,
            },
            {
                "performative": SigningMessage.Performative.SIGNED_TRANSACTION,
            },
            {
                "performative": LedgerApiMessage.Performative.TRANSACTION_DIGEST,
            },
            {
                "performative": LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
            },
        ]
        expected_types_enter: ExpectedTypesType = [
            {
                "raw_transaction": RawTransaction,
            },
            {
                "signed_transaction": SignedTransaction,
            },
            {
                "transaction_digest": TransactionDigest,
            },
            {
                "transaction_receipt": TransactionReceipt,
            },
        ]
        _, _, _, msg4 = self.process_n_messages(
            cycles_enter,
            self.price_estimation_synchronized_data,
            DeployOracleBehaviour.behaviour_id,
            handlers_enter,
            expected_content_enter,
            expected_types_enter,
        )
        assert msg4 is not None and isinstance(msg4, LedgerApiMessage)
        oracle_contract_address = EthereumApi.get_contract_address(
            msg4.transaction_receipt.receipt
        )

        # update synchronized data with oracle contract address
        self.price_estimation_synchronized_data.update(
            oracle_contract_address=oracle_contract_address,
        )

    def gen_safe_tx_hash(self) -> None:
        """Generate safe's transaction hash."""
        cycles_enter = 3
        handlers_enter: HandlersType = [self.contract_handler] * cycles_enter
        expected_content_enter: ExpectedContentType = [
            {"performative": ContractApiMessage.Performative.RAW_TRANSACTION}
        ] * cycles_enter
        expected_types_enter: ExpectedTypesType = [
            {
                "raw_transaction": RawTransaction,
            }
        ] * cycles_enter
        _, msg_a, msg_b = self.process_n_messages(
            cycles_enter,
            self.price_estimation_synchronized_data,
            TransactionHashBehaviour.behaviour_id,
            handlers_enter,
            expected_content_enter,
            expected_types_enter,
        )

        assert msg_a is not None and isinstance(msg_a, ContractApiMessage)
        tx_data = cast(bytes, msg_a.raw_transaction.body["data"])
        assert msg_b is not None and isinstance(msg_b, ContractApiMessage)
        tx_hash = cast(str, msg_b.raw_transaction.body["tx_hash"])[2:]

        payload = hash_payload_to_hex(
            tx_hash,
            ETHER_VALUE,
            SAFE_TX_GAS,
            self.price_estimation_synchronized_data.oracle_contract_address,
            tx_data,
        )

        # update synchronized data with safe's tx hash
        self.tx_settlement_synchronized_data.update(
            most_voted_tx_hash=payload,
        )

    def clear_unmined_txs(self) -> None:
        """Clear all unmined txs. Mined txs will not be cleared, but this is not a problem."""
        for tx in self.tx_settlement_synchronized_data.tx_hashes_history:
            self.hardhat_provider.make_request(
                RPCEndpoint("hardhat_dropTransaction"), (tx,)
            )

    @staticmethod
    def dummy_try_get_gas_pricing_wrapper(
        max_priority_fee_per_gas: Wei = DUMMY_MAX_PRIORITY_FEE_PER_GAS,
        max_fee_per_gas: Wei = DUMMY_MAX_FEE_PER_GAS,
        repricing_multiplier: float = DUMMY_REPRICING_MULTIPLIER,
    ) -> Callable[
        [Optional[str], Optional[Dict], Optional[Dict[str, Wei]]], Dict[str, Wei]
    ]:
        """A dummy wrapper of `EthereumAPI`'s `try_get_gas_pricing`."""

        def dummy_try_get_gas_pricing(
            _gas_price_strategy: Optional[str] = None,
            _extra_config: Optional[Dict] = None,
            old_price: Optional[Dict[str, Wei]] = None,
        ) -> Dict[str, Wei]:
            """Get a dummy gas price."""
            tip = max_priority_fee_per_gas
            gas = max_fee_per_gas

            if old_price is not None:
                tip = ceil(max_priority_fee_per_gas * repricing_multiplier)
                gas = ceil(max_fee_per_gas * repricing_multiplier)
            return {"maxPriorityFeePerGas": tip, "maxFeePerGas": gas}

        return dummy_try_get_gas_pricing


class TestRepricing(TransactionSettlementIntegrationBaseCase):
    """Test failure modes related to repricing."""

    @pytest.mark.parametrize("should_mock_ledger_pricing_mechanism", (True, False))
    def test_same_keeper(
        self,
        should_mock_ledger_pricing_mechanism: bool,
    ) -> None:
        """Test repricing with and without mocking ledger's `try_get_gas_pricing` method."""

        try:
            if should_mock_ledger_pricing_mechanism:
                with mock.patch.object(
                    EthereumApi,
                    "try_get_gas_pricing",
                    new_callable=TransactionSettlementIntegrationBaseCase.dummy_try_get_gas_pricing_wrapper,
                ):
                    self._test_same_keeper()
            else:
                self._test_same_keeper()

        finally:
            self.clear_unmined_txs()

    def _test_same_keeper(self) -> None:
        """
        Test repricing after the first failure.

        Test that we are using the same keeper to reprice when we fail or timeout for the first time.
        Also, test that we are adjusting the gas correctly when repricing.
        """

        # deploy the oracle
        self.deploy_oracle()
        # generate tx hash
        self.gen_safe_tx_hash()
        # sign tx
        self.sign_tx()
        # stop HardHat's automatic mining
        assert self.hardhat_provider.make_request(
            RPCEndpoint("evm_setAutomine"), [False]
        ), "Disabling auto-mining failed!"
        # send tx first time, we expect it to be pending until we enable the mining back
        self.send_tx()
        # re-enable HardHat's automatic mining so that the second tx replaces the first, pending one
        assert self.hardhat_provider.make_request(
            RPCEndpoint("evm_setIntervalMining"), [1000]
        ), "Re-enabling auto-mining failed!"
        # send tx second time
        self.send_tx()
        # validate the tx
        self.validate_tx()


class TestKeepers(OracleBehaviourBaseCase, IntegrationBaseCase):
    """Test the keepers related functionality for the tx settlement skill."""

    @classmethod
    def setup(cls, **kwargs: Any) -> None:
        """Set up the test class."""
        super().setup()

        # init synchronized data
        cls.tx_settlement_synchronized_data = TxSettlementSynchronizedSata(
            AbciAppDB(
                initial_data=dict(
                    participants=[frozenset(list(cls.agents.keys()))],
                    most_voted_randomness=["0xabcd"],
                ),
            )
        )

    @mock.patch.object(
        TransactionSettlementBaseBehaviour,
        "serialized_keepers",
        side_effect=lambda keepers, retries: retries.to_bytes(32, "big").hex()
        + "".join(keepers),
    )
    def select_keeper(
        self,
        serialized_keepers_mock: mock.Mock,
        expected_keepers: Deque[str],
        expected_retries: int,
        first_time: bool = False,
    ) -> None:
        """Select a keeper."""

        if first_time:
            behaviour_id = SelectKeeperTransactionSubmissionBehaviourA.behaviour_id
        else:
            behaviour_id = SelectKeeperTransactionSubmissionBehaviourB.behaviour_id

        # select keeper
        self.fast_forward_to_behaviour(
            self.behaviour,
            behaviour_id,
            self.tx_settlement_synchronized_data,
        )
        assert self.behaviour.current_behaviour is not None
        assert self.behaviour.current_behaviour.behaviour_id == behaviour_id

        self.behaviour.act_wrapper()
        serialized_keepers_mock.assert_called_with(expected_keepers, expected_retries)

        # update keepers.
        self.tx_settlement_synchronized_data.update(
            # we cast to A class, because it is the top level one between A & B, and we need `serialized_keepers`
            keepers=cast(
                SelectKeeperTransactionSubmissionBehaviourA,
                self.behaviour.current_behaviour,
            ).serialized_keepers(expected_keepers, expected_retries)
        )

    def test_keepers_alternating(self) -> None:
        """Test that we are alternating the keepers when we fail or timeout more than `keeper_allowed_retries` times."""
        # set verification status
        self.tx_settlement_synchronized_data.update(
            final_verification_status=VerificationStatus.PENDING,
        )

        # select keeper a
        self.select_keeper(
            expected_keepers=deque(("0xBcd4042DE499D14e55001CcbB24a551F3b954096",)),
            expected_retries=1,
            first_time=True,
        )
        assert isinstance(
            self.behaviour.current_behaviour,
            SelectKeeperTransactionSubmissionBehaviourA,
        )

        for i in range(
            self.behaviour.current_behaviour.params.keeper_allowed_retries - 1
        ):
            # select keeper b
            # ensure that we select the same keeper until the `keeper_allowed_retries` is reached.
            # +2 because we selected once for keeperA and also index starts from 0.
            self.select_keeper(
                expected_keepers=deque(("0xBcd4042DE499D14e55001CcbB24a551F3b954096",)),
                expected_retries=i + 2,
            )

        # select keeper b after retries are reached.
        self.select_keeper(
            expected_keepers=deque(
                (
                    "0x71bE63f3384f5fb98995898A86B02Fb2426c5788",
                    "0xBcd4042DE499D14e55001CcbB24a551F3b954096",
                )
            ),
            expected_retries=1,
        )

    def test_rotation(self) -> None:
        """Test keepers rotating when threshold reached."""
        # set more keepers than the threshold
        keepers = deque(
            (
                "0xBcd4042DE499D14e55001CcbB24a551F3b954096",
                "0x71bE63f3384f5fb98995898A86B02Fb2426c5788",
            )
        )
        keeper_retries = 1

        self.tx_settlement_synchronized_data.update(
            keepers=int(keeper_retries).to_bytes(32, "big").hex() + "".join(keepers),
        )

        expected_keepers = keepers.copy()
        # test twice as many times as the number of participants
        for _ in range(len(self.tx_settlement_synchronized_data.participants) * 2):
            # rotate expected keepers
            expected_keepers.rotate(-1)
            # select keeper b
            self.select_keeper(expected_keepers=expected_keepers, expected_retries=1)


class TestSyncing(TransactionSettlementIntegrationBaseCase):
    """Test late tx hashes synchronization."""

    @classmethod
    def setup(cls, **kwargs: Any) -> None:
        """Set up the test class."""
        super().setup()

        # update synchronized data
        cls.tx_settlement_synchronized_data.update(missed_messages=0)

    def sync_late_messages(self) -> None:
        """Synchronize late messages."""
        params = cast(
            TransactionSettlementBaseBehaviour, self.behaviour.current_behaviour
        ).params
        late_messages_len = len(params.late_messages)
        expected_sync_result = params.tx_hash

        handlers: HandlersType = [
            self.signing_handler,
            self.ledger_handler,
        ] * late_messages_len
        expected_content: ExpectedContentType = [
            {"performative": SigningMessage.Performative.SIGNED_MESSAGE},
            {"performative": LedgerApiMessage.Performative.TRANSACTION_DIGEST},
        ] * late_messages_len
        expected_types: ExpectedTypesType = [
            {
                "signed_message": SignedMessage,
            },
            {
                "transaction_digest": TransactionDigest,
            },
        ] * late_messages_len
        msgs = self.process_n_messages(
            len(handlers),
            self.tx_settlement_synchronized_data,
            SynchronizeLateMessagesBehaviour.behaviour_id,
            handlers,
            expected_content,
            expected_types,
        )

        assert isinstance(
            self.behaviour.current_behaviour, SynchronizeLateMessagesBehaviour
        )
        assert (
            self.behaviour.current_behaviour.behaviour_id
            == SynchronizeLateMessagesBehaviour.behaviour_id
        )
        assert self.behaviour.current_behaviour.params.tx_hash == ""
        assert self.behaviour.current_behaviour.params.late_messages == []

        tx_digest_msgs = msgs[0::2]
        for i in range(len(tx_digest_msgs)):
            current_message = tx_digest_msgs[i]
            assert current_message is not None and isinstance(
                current_message, LedgerApiMessage
            )
            tx_digest = current_message.transaction_digest.body
            assert isinstance(tx_digest, str)
            assert (
                tx_digest
            ), f"No tx digest retrieved for message {i}: {current_message}!"
            expected_sync_result += tx_digest

        assert self.behaviour.current_behaviour._tx_hashes == expected_sync_result
        self.tx_settlement_synchronized_data.update(
            late_arriving_tx_hashes=[expected_sync_result],
        )
        self.tx_settlement_synchronized_data.update(
            missed_messages=self.behaviour.current_behaviour.synchronized_data.missed_messages
            - len(
                self.behaviour.current_behaviour.synchronized_data.late_arriving_tx_hashes
            ),
        )

    def check_late_tx_hashes(self) -> None:
        """Check the late transaction hashes to see if any is validated."""
        handlers: HandlersType = [
            self.contract_handler,
        ] * len(self.tx_settlement_synchronized_data.late_arriving_tx_hashes)
        expected_content: ExpectedContentType = [
            {"performative": ContractApiMessage.Performative.STATE},
        ] * len(self.tx_settlement_synchronized_data.late_arriving_tx_hashes)
        expected_types: ExpectedTypesType = [
            {
                "state": State,
            },
        ] * len(self.tx_settlement_synchronized_data.late_arriving_tx_hashes)
        msgs = self.process_n_messages(
            len(self.tx_settlement_synchronized_data.late_arriving_tx_hashes),
            self.tx_settlement_synchronized_data,
            CheckLateTxHashesBehaviour.behaviour_id,
            handlers,
            expected_content,
            expected_types,
        )
        assert isinstance(self.behaviour.current_behaviour, CheckLateTxHashesBehaviour)
        assert (
            self.behaviour.current_behaviour.behaviour_id
            == CheckLateTxHashesBehaviour.behaviour_id
        )

        verified_idx = -1
        verified_count = 0
        for i in range(len(msgs)):
            current_message = msgs[i]
            assert current_message is not None and isinstance(
                current_message, ContractApiMessage
            )
            if current_message.state.body["verified"]:
                verified_idx = i
                verified_count += 1
        assert verified_idx != -1, f"No message has been verified: {msgs}"
        assert verified_count == 1, "More than 1 messages have been verified!"

        self.tx_settlement_synchronized_data.update(
            final_verification_status=VerificationStatus.VERIFIED,
            final_tx_hash=self.tx_settlement_synchronized_data.late_arriving_tx_hashes[
                -verified_idx
            ],
        )

    @mock.patch.object(
        EthereumApi,
        "try_get_gas_pricing",
        side_effect=TransactionSettlementIntegrationBaseCase.dummy_try_get_gas_pricing_wrapper(),
    )
    def test_sync_local_hash(self, _: mock.Mock) -> None:
        """Test the case in which we have received a tx hash during finalization, but timed out before sharing it."""
        # deploy the oracle
        self.deploy_oracle()
        # generate tx hash
        self.gen_safe_tx_hash()
        # sign tx
        self.sign_tx()
        # send tx, but do not update the state in order to simulate a round's time out.
        self.send_tx(simulate_timeout=True)
        # check that we have increased the number of missed messages.
        assert self.tx_settlement_synchronized_data.missed_messages == 1
        # store the tx hash that we have missed.
        assert isinstance(self.behaviour.current_behaviour, FinalizeBehaviour)
        missed_hash = self.behaviour.current_behaviour.params.tx_hash
        # sync the tx hash that we missed before
        self.sync_late_messages()
        # check that we have decreased the number of missed messages.
        assert self.tx_settlement_synchronized_data.missed_messages == 0
        # check the tx hash that we missed before to see if it is verified
        self.check_late_tx_hashes()
        assert (
            self.tx_settlement_synchronized_data.final_verification_status
            == VerificationStatus.VERIFIED
        )
        assert self.tx_settlement_synchronized_data.final_tx_hash == missed_hash



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains all the storing operations of the behaviours."""


import json
import os.path
from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Any, Callable, Dict, Optional, TypeVar, Union, cast

import joblib
import pandas as pd
from pmdarima.pipeline import Pipeline

from packages.valory.skills.abstract_round_abci.io.paths import create_pathdirs


StoredJSONType = Union[dict, list]
NativelySupportedSingleObjectType = Union[StoredJSONType, Pipeline, pd.DataFrame]
NativelySupportedMultipleObjectsType = Dict[str, NativelySupportedSingleObjectType]
NativelySupportedObjectType = Union[
    NativelySupportedSingleObjectType, NativelySupportedMultipleObjectsType
]
NativelySupportedStorerType = Callable[[str, NativelySupportedObjectType, Any], None]
CustomObjectType = TypeVar("CustomObjectType")
CustomStorerType = Callable[[str, CustomObjectType, Any], None]
SupportedSingleObjectType = Union[NativelySupportedObjectType, CustomObjectType]
SupportedMultipleObjectsType = Dict[str, SupportedSingleObjectType]
SupportedObjectType = Union[SupportedSingleObjectType, SupportedMultipleObjectsType]
SupportedStorerType = Union[NativelySupportedStorerType, CustomStorerType]
NativelySupportedJSONStorerType = Callable[
    [str, Union[StoredJSONType, Dict[str, StoredJSONType]], Any], None
]
NativelySupportedPipelineStorerType = Callable[
    [str, Union[Pipeline, Dict[str, Pipeline]], Any], None
]
NativelySupportedDfStorerType = Callable[
    [str, Union[pd.DataFrame, Dict[str, pd.DataFrame]], Any], None
]


class SupportedFiletype(Enum):
    """Enum for the supported filetypes of the IPFS interacting methods."""

    JSON = auto()
    PM_PIPELINE = auto()
    CSV = auto()


class AbstractStorer(ABC):
    """An abstract `Storer` class."""

    def __init__(self, path: str):
        """Initialize an abstract storer."""
        self._path = path
        # Create the dirs of the path if it does not exist.
        create_pathdirs(path)

    @abstractmethod
    def store_single_file(
        self, filename: str, obj: SupportedSingleObjectType, **kwargs: Any
    ) -> None:
        """Store a single file."""

    def store(self, obj: SupportedObjectType, multiple: bool, **kwargs: Any) -> None:
        """Store one or multiple files."""
        if multiple:
            if not isinstance(obj, dict):  # pragma: no cover
                raise ValueError(
                    f"Cannot store multiple files of type {type(obj)}!"
                    f"Should be a dictionary of filenames mapped to their objects."
                )
            for filename, single_obj in obj.items():
                filename = os.path.join(self._path, filename)
                self.store_single_file(filename, single_obj, **kwargs)
        else:
            self.store_single_file(self._path, obj, **kwargs)


class JSONStorer(AbstractStorer):
    """A JSON file storer."""

    def store_single_file(
        self, filename: str, obj: NativelySupportedSingleObjectType, **kwargs: Any
    ) -> None:
        """Store a JSON."""
        if not any(isinstance(obj, type_) for type_ in (dict, list)):
            raise ValueError(  # pragma: no cover
                f"`JSONStorer` cannot be used with a {type(obj)}! Only with a {StoredJSONType}"
            )

        try:
            with open(filename, "w", encoding="utf-8") as f:
                json.dump(obj, f, ensure_ascii=False, indent=4)
        except (TypeError, OSError) as e:  # pragma: no cover
            raise IOError(str(e)) from e


class CSVStorer(AbstractStorer):
    """A CSV file storer."""

    def store_single_file(
        self, filename: str, obj: NativelySupportedSingleObjectType, **kwargs: Any
    ) -> None:
        """Store a pandas dataframe."""
        if not isinstance(obj, pd.DataFrame):
            raise ValueError(  # pragma: no cover
                f"`JSONStorer` cannot be used with a {type(obj)}! Only with a {pd.DataFrame}"
            )

        index = kwargs.get("index", False)

        try:
            obj.to_csv(filename, index=index)
        except (TypeError, OSError) as e:  # pragma: no cover
            raise IOError(str(e)) from e


class ForecasterStorer(AbstractStorer):
    """A pmdarima Pipeline storer."""

    def store_single_file(
        self, filename: str, obj: NativelySupportedSingleObjectType, **kwargs: Any
    ) -> None:
        """Store a pmdarima Pipeline."""
        if not isinstance(obj, Pipeline):
            raise ValueError(  # pragma: no cover
                f"`JSONStorer` cannot be used with a {type(obj)}! Only with a {Pipeline}"
            )

        try:
            joblib.dump(obj, filename)
        except (ValueError, OSError) as e:  # pragma: no cover
            raise IOError(str(e)) from e


class Storer(AbstractStorer):
    """Class which stores files."""

    def __init__(
        self,
        filetype: Optional[SupportedFiletype],
        custom_storer: Optional[CustomStorerType],
        path: str,
    ):
        """Initialize a `Storer`."""
        super().__init__(path)
        self._filetype = filetype
        self._custom_storer = custom_storer
        self.__filetype_to_storer: Dict[SupportedFiletype, SupportedStorerType] = {
            SupportedFiletype.JSON: cast(
                NativelySupportedJSONStorerType, JSONStorer(path).store_single_file
            ),
            SupportedFiletype.PM_PIPELINE: cast(
                NativelySupportedPipelineStorerType,
                ForecasterStorer(path).store_single_file,
            ),
            SupportedFiletype.CSV: cast(
                NativelySupportedDfStorerType, CSVStorer(path).store_single_file
            ),
        }

    def store_single_file(
        self, filename: str, obj: NativelySupportedObjectType, **kwargs: Any
    ) -> None:
        """Store a single file."""
        storer = self._get_single_storer_from_filetype()
        storer(filename, obj, **kwargs)  # type: ignore

    def _get_single_storer_from_filetype(self) -> SupportedStorerType:
        """Get a file storer from a given filetype or keep a custom storer."""
        if self._filetype is not None:
            return self.__filetype_to_storer[self._filetype]

        if self._custom_storer is not None:  # pragma: no cover
            return self._custom_storer

        raise ValueError(  # pragma: no cover
            "Please provide either a supported filetype or a custom storing function."
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""ETL related operations."""

from typing import Callable, Dict, List, Optional, Union, cast

import pandas as pd


ResponseItemType = List[Dict[str, Union[str, Dict[str, str]]]]

# Define a dictionary with the data types for each column of the historical data.
HIST_DTYPES = {
    "createdAtBlockNumber": int,
    "createdAtTimestamp": int,
    "id": str,
    "liquidityProviderCount": int,
    "reserve0": float,
    "reserve1": float,
    "reserveETH": float,
    "reserveUSD": float,
    "token0Price": float,
    "token1Price": float,
    "totalSupply": float,
    "trackedReserveETH": float,
    "untrackedVolumeUSD": float,
    "txCount": int,
    "volumeToken0": float,
    "volumeToken1": float,
    "volumeUSD": float,
    "forTimestamp": int,
    "blockNumber": int,
    "blockTimestamp": int,
    "ethPrice": float,
    "token0": object,
    "token1": object,
}

# Define a dictionary with the data types of the columns of the transformed historical data.
TRANSFORMED_HIST_DTYPES = HIST_DTYPES.copy()

NEW_STR_COLS = (
    "token0ID",
    "token0Name",
    "token0Symbol",
    "token1ID",
    "token1Name",
    "token1Symbol",
    "pairName",
)

NEW_FLOAT_COLS = ("updatedVolumeUSD", "updatedReserveUSD", "currentChange", "APY")

TOKEN_COL_NAMES = ("token0", "token1")

for new_str_col in NEW_STR_COLS:
    TRANSFORMED_HIST_DTYPES[new_str_col] = str

for new_float_col in NEW_FLOAT_COLS:
    TRANSFORMED_HIST_DTYPES[new_float_col] = float

for old_col in TOKEN_COL_NAMES:
    del TRANSFORMED_HIST_DTYPES[old_col]


def calc_change(volume_usd: pd.Series) -> pd.Series:
    """Calculate the change between two days. The series argument needs to be sorted by date!

    :param volume_usd: a series with volumes, sorted by date.
    :return: a new series with the current day's change for each date of the given `volume_usd` series.
    """
    return volume_usd - volume_usd.shift(1)


def calc_apy(x: pd.Series) -> Optional[float]:
    """Calculates the APY for a given pool's info.

    :param x: a series with historical data.
    :return: the daily APY.
    """
    res = None

    if x["updatedReserveUSD"]:
        res = (x["currentChange"] * 0.002 * 365 * 100) / x["updatedReserveUSD"]

    return res


def apply_hist_based_calculations(pairs_hist: pd.DataFrame) -> None:
    """Fill the given pairs' history with the current change in USD and APY values."""
    # Calculate the current change of volume in USD.
    pairs_hist["currentChange"] = pairs_hist.groupby("id")[
        "updatedVolumeUSD"
    ].transform(calc_change)

    # Drop NaN values (essentially, this is the first day's `current_change`, because we cannot calculate it).
    pairs_hist.dropna(subset=["currentChange"], inplace=True)

    if len(pairs_hist.index) == 0:
        raise ValueError(
            "APY cannot be calculated if there are not at least two observations for a pool!"
        )

    # Calculate APY.
    pairs_hist["APY"] = pairs_hist.apply(calc_apy, axis=1)
    # Drop rows with NaN APY values.
    pairs_hist.dropna(subset=["APY"], inplace=True)


def transform_hist_data(
    pairs_hist_raw: ResponseItemType, batch: bool = False
) -> pd.DataFrame:
    """Transform pairs' history into a dataframe and add extra fields.

    :param pairs_hist_raw: the pairs historical data non-transformed.
    :param batch: whether the input is a batch which contains single a data point per pool.
    :return: a dataframe with the given historical data, containing extra fields. These are:
         * [token0ID, token0Name, token0Symbol]: split from `token0`.
         * [token1ID, token1Name, token1Symbol]: split from `token1`.
         * pairName: the current's pair's name, which is derived by: 'token0_name - token1_name'.
         * updatedVolumeUSD: the tracked volume USD, but if it is 0, then the untracked volume USD.
         * updatedReserveUSD: the tracked reserve USD, but if it is 0, then the untracked reserve USD.
         * current_change: the current day's change in volume USD, only if `batch` is `False`.
         * APY: the APY, only if `batch` is `False`.

         The dataframe is also sorted by the block's timestamp and the pair's tokens' names,
         and the entries for which the APY cannot be calculated are being dropped.
    """
    # Convert history to a dataframe.
    pairs_hist = pd.DataFrame(pairs_hist_raw).astype(HIST_DTYPES)

    # Split the dictionary-like token cols.
    for token_col in TOKEN_COL_NAMES:
        pairs_hist[f"{token_col}ID"] = (
            pairs_hist[token_col].apply(lambda x: x["id"]).astype(str)
        )
        pairs_hist[f"{token_col}Name"] = (
            pairs_hist[token_col].apply(lambda x: x["name"]).astype(str)
        )
        pairs_hist[f"{token_col}Symbol"] = (
            pairs_hist[token_col].apply(lambda x: x["symbol"]).astype(str)
        )
    # Drop the original dictionary-like token cols.
    pairs_hist.drop(columns=list(TOKEN_COL_NAMES), inplace=True)

    # Create pair's name.
    pairs_hist["pairName"] = pairs_hist["token0Name"] + " - " + pairs_hist["token1Name"]

    # Create a volume USD and a reserve USD taking the untracked amount into consideration as well.
    pairs_hist["updatedVolumeUSD"] = pairs_hist["volumeUSD"]
    pairs_hist["updatedReserveUSD"] = (
        pairs_hist["trackedReserveETH"] * pairs_hist["ethPrice"]
    )

    # Create a mask for all the untracked cases.
    not_tracked_m = pairs_hist["updatedVolumeUSD"] == 0

    # Keep the untracked amount for the untracked cases.
    pairs_hist.loc[not_tracked_m, "updatedVolumeUSD"] = pairs_hist.loc[
        not_tracked_m, "untrackedVolumeUSD"
    ]
    pairs_hist.loc[not_tracked_m, "updatedReserveUSD"] = pairs_hist.loc[
        not_tracked_m, "reserveUSD"
    ]

    if not batch:
        apply_hist_based_calculations(pairs_hist)

    # Sort the dictionary.
    pairs_hist.sort_values(
        by=["blockTimestamp", "token0Symbol", "token1Symbol"],
        ascending=True,
        inplace=True,
    )

    return pairs_hist


def prepare_batch(
    previous_batch: pd.DataFrame, current_batch_raw: ResponseItemType
) -> pd.DataFrame:
    """Prepare a batch, using the currently fetched batch from the subgraph and the last utilized batch.

    :param previous_batch: the last utilized batch.
    :param current_batch_raw: the currently fetched data, non-transformed.
    :return: a dictionary with the pool ids, mapped to their current data point of the timeseries.
    """
    # Transform the current batch.
    current_batch = transform_hist_data(current_batch_raw, batch=True)
    # Append the current batch to the previous batch.
    batches = pd.concat([previous_batch, current_batch])
    # Calculate the last APY value per pool, using the batches.
    prepared_batches = []
    for pool_id, pool_batch in batches.groupby("id"):
        if len(pool_batch.index) < 2:
            raise ValueError(
                f"Could not find any previous history in {pool_batch} for pool `{pool_id}`!"
            )
        # Since we have concatenated the current batch after the previous batch
        # and `groupby` preserves the order of rows within each group,
        # then we do not need to worry about the sorting of the batches.
        apply_hist_based_calculations(pool_batch)
        prepared_batches.append(pool_batch)

    return pd.concat(prepared_batches)


def apply_revert_token_cols_wrapper(
    token_name: str,
) -> Callable[[pd.Series], Dict[str, str]]:
    """A wrapper to revert a token column as it was before the transformation.

    :param token_name: the token name for which we want to revert.
    :return: a dictionary of the initial representation.
    """

    def apply_revert_token_cols(x: pd.Series) -> Dict[str, str]:
        """Revert the token column as it was before the transformation.

        :param x: the input series row.
        :return: a dictionary of the initial representation.
        """
        return {
            "id": x[f"{token_name}ID"],
            "name": x[f"{token_name}Name"],
            "symbol": x[f"{token_name}Symbol"],
        }

    return apply_revert_token_cols


def revert_transform_hist_data(pairs_hist: pd.DataFrame) -> ResponseItemType:
    """Revert the transformation of pairs' history.

    :param pairs_hist: the transformed pairs' historical data.
    :return: the reverted historical data.
    """
    # Re-create the old data which have been dropped.
    for token_name in TOKEN_COL_NAMES:
        pairs_hist[token_name] = pairs_hist.apply(
            apply_revert_token_cols_wrapper(token_name), axis=1
        )

    # Drop columns that occur after the transformation.
    drop_cols = list(NEW_STR_COLS + NEW_FLOAT_COLS)
    for token_name in TOKEN_COL_NAMES:
        drop_cols.extend(
            [f"{token_name}ID", f"{token_name}Name", f"{token_name}Symbol"]
        )
    pairs_hist.drop(columns=drop_cols, inplace=True)

    # Convert timestamp to unix int.
    pairs_hist["blockTimestamp"] = pairs_hist["blockTimestamp"].view(int) / 10 ** 9

    # Convert history to a list of dicts.
    reverted_pairs_hist = pairs_hist.to_dict("records")

    return cast(ResponseItemType, reverted_pairs_hist)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Deploy CLI module."""

import os
import shutil
from pathlib import Path
from typing import Optional

import click
from aea.cli.utils.click_utils import PublicIdParameter, password_option
from aea.configurations.constants import PACKAGES
from aea.configurations.data_types import PublicId

from autonomy.cli.utils.click_utils import image_profile_flag
from autonomy.constants import DEFAULT_IMAGE_VERSION
from autonomy.deploy.build import generate_deployment
from autonomy.deploy.generators.docker_compose.base import DockerComposeGenerator
from autonomy.deploy.generators.kubernetes.base import KubernetesGenerator
from autonomy.deploy.image import ImageBuilder


@click.group(name="deploy")
def deploy_group() -> None:
    """Deploy an AEA project."""


@deploy_group.group(name="build")
def build_group() -> None:
    """Build tools"""


@build_group.command(name="deployment")
@click.argument(
    "service-id",
    type=PublicIdParameter(),
)
@click.argument("keys_file", type=str, required=True)
@click.option(
    "--o",
    "output_dir",
    type=click.Path(exists=False, dir_okay=True),
    default=Path.cwd(),
    help="Path to output dir.",
)
@click.option(
    "--n",
    "number_of_agents",
    type=int,
    default=None,
    help="Number of agents.",
)
@click.option(
    "--docker",
    "deployment_type",
    flag_value=DockerComposeGenerator.deployment_type,
    default=True,
    help="Use docker as a backend.",
)
@click.option(
    "--kubernetes",
    "deployment_type",
    flag_value=KubernetesGenerator.deployment_type,
    help="Use docker as a kubernetes.",
)
@click.option(
    "--packages-dir",
    type=click.Path(exists=True, dir_okay=True),
    default=Path.cwd() / PACKAGES,
    help="Path to packages folder (For local usage).",
)
@click.option(
    "--dev",
    "dev_mode",
    is_flag=True,
    default=False,
    help="Create development environment.",
)
@click.option(
    "--version",
    "version",
    help="Specify deployment version.",
)
@click.option(
    "--force",
    "force_overwrite",
    is_flag=True,
    default=False,
    help="Remove existing build and overwrite with new one.",
)
@password_option(confirmation_prompt=True)
def build_deployment(  # pylint: disable=too-many-arguments
    service_id: PublicId,
    keys_file: Path,
    deployment_type: str,
    output_dir: Path,
    packages_dir: Path,
    dev_mode: bool,
    force_overwrite: bool,
    number_of_agents: Optional[int] = None,
    password: Optional[str] = None,
    version: Optional[str] = None,
) -> None:
    """Build deployment setup for n agents."""

    packages_dir = Path(packages_dir)
    keys_file = Path(keys_file)
    build_dir = Path(output_dir, "abci_build")

    if not packages_dir.is_dir():
        raise click.ClickException(
            f"Packages directory does not exists @ {packages_dir}"
        )

    if build_dir.is_dir():
        if not force_overwrite:
            raise click.ClickException(f"Build already exists @ {output_dir}")
        shutil.rmtree(build_dir)

    build_dir.mkdir()
    _build_dirs(build_dir)

    service_path = _find_path_to_service(service_id, packages_dir)
    try:
        report = generate_deployment(
            service_path=service_path,
            type_of_deployment=deployment_type,
            private_keys_file_path=keys_file,
            private_keys_password=password,
            number_of_agents=number_of_agents,
            packages_dir=packages_dir,
            build_dir=build_dir,
            dev_mode=dev_mode,
            version=version,
        )
        click.echo(report)
    except Exception as e:  # pylint: disable=broad-except
        raise click.ClickException(str(e)) from e


@build_group.command(name="image")
@click.argument(
    "service-id",
    type=PublicIdParameter(),
)
@click.option(
    "--packages-dir",
    type=click.Path(exists=True, dir_okay=True),
    default=Path.cwd() / PACKAGES,
    help="Path to packages folder (For local usage).",
)
@click.option(
    "--build-dir",
    type=click.Path(exists=True, dir_okay=True),
    default=Path.cwd() / "deployments" / "Dockerfiles" / "open_aea",
    help="Path to build directory.",
)
@click.option(
    "--skaffold-dir",
    type=click.Path(exists=True, dir_okay=True),
    default=Path.cwd(),
    help="Path to directory containing the skaffold config.",
)
@click.option(
    "--version",
    type=str,
    default=DEFAULT_IMAGE_VERSION,
    help="Image version",
)
@click.option("--push", is_flag=True, default=False, help="Push image after build.")
@image_profile_flag()
def build_images(  # pylint: disable=too-many-arguments
    service_id: PublicId,
    profile: str,
    packages_dir: Path,
    build_dir: Path,
    skaffold_dir: Path,
    version: str,
    push: bool,
) -> None:
    """Build image using skaffold."""

    packages_dir = Path(packages_dir).absolute()
    build_dir = Path(build_dir).absolute()
    skaffold_dir = Path(skaffold_dir).absolute()

    try:
        click.echo(
            f"Building image with:\n\tProfile: {profile}\n\tServiceId: {service_id}\n"
        )
        deployment_file_path = _find_path_to_service_file(service_id, packages_dir)
        ImageBuilder.build_images(
            profile=profile,
            deployment_file_path=deployment_file_path,
            push=push,
            packages_dir=packages_dir,
            build_dir=build_dir,
            version=version,
            skaffold_dir=skaffold_dir,
        )
    except Exception as e:  # pylint: disable=broad-except
        raise click.ClickException(str(e)) from e


def _find_path_to_service_file(public_id: PublicId, packages_dir: Path) -> Path:
    """Find path to service file using package dir."""
    service_file = (
        packages_dir / public_id.author / "services" / public_id.name / "service.yaml"
    )
    if not service_file.is_file():
        raise click.ClickException(f"Cannot find service file for {public_id}")

    return service_file


def _find_path_to_service(public_id: PublicId, packages_dir: Path) -> Path:
    """Find path to service file using package dir."""
    service_path = packages_dir / public_id.author / "services" / public_id.name
    if not service_path.is_dir():
        raise click.ClickException(f"Cannot find service file for {public_id}")

    return service_path


def _build_dirs(build_dir: Path) -> None:
    """Build necessary directories."""

    for dir_path in [
        ("persistent_data",),
        ("persistent_data", "logs"),
        ("persistent_data", "tm_state"),
        ("persistent_data", "benchmarks"),
        ("persistent_data", "venvs"),
        ("agent_keys",),
    ]:
        path = Path(build_dir, *dir_path)
        path.mkdir()
        # TOFIX for macOS
        try:
            os.chown(path, 1000, 1000)
        except PermissionError:
            click.echo(
                f"Updating permissions failed for {path}, please do it manually."
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Kubernetes Templates module."""


HARDHAT_TEMPLATE: str = """apiVersion: apps/v1
kind: Deployment
metadata:
  labels:
    io.kompose.service: hardhat
  name: hardhat
spec:
  replicas: 1
  selector:
    matchLabels:
      io.kompose.service: hardhat
  strategy:
    type: Recreate
  template:
    metadata:
      labels:
        io.kompose.network/localnet: "true"
        io.kompose.service: hardhat
    spec:
      containers:
        - args:
            - /usr/local/bin/yarn
            - run
            - hardhat
            - node
            - --port
            - "8545"
            - --hostname
            - "0.0.0.0"
          command:
            - /bin/bash
          image: %s:%s
          name: hardhat
          ports:
            - name: http
              containerPort: 8545
          resources: {}
          workingDir: /home/ubuntu/build
          imagePullPolicy: Always
      imagePullSecrets:
      - name: regcred
      restartPolicy: Always
status: {}
---
apiVersion: v1
kind: Service
metadata:
  labels:
    io.kompose.service: hardhat
  name: hardhat
spec:
  ports:
    - protocol: TCP
      port: 8545
      targetPort: 8545
  selector:
    io.kompose.service: hardhat
status:
  loadBalancer: {}
"""


CLUSTER_CONFIGURATION_TEMPLATE: str = """apiVersion: batch/v1
kind: Job
metadata:
  name: config-nodes
spec:
  template:
    spec:
      imagePullSecrets:
      - name: regcred
      containers:
      - name: config-nodes
        image: {tendermint_image_name}:{tendermint_image_version}
        command: ['/usr/bin/tendermint']
        args: ["testnet",
         "--config",
         "/etc/tendermint/config-template.toml",
         "--o",  ".", {host_names},
         "--v", "{number_of_validators}"
         ]
        volumeMounts:
          - mountPath: /tendermint
            name: nodes
      volumes:
        - name: nodes
          persistentVolumeClaim:
            claimName: 'nodes'
      restartPolicy: Never
  backoffLimit: 3
---
apiVersion: v1
kind: PersistentVolumeClaim
metadata:
  name: logs-pvc
spec:
  storageClassName: nfs-client
  accessModes:
    - ReadWriteOnce
  resources:
    requests:
      storage: 1000M
---
apiVersion: v1
kind: PersistentVolumeClaim
metadata:
  name: tendermint-pvc
spec:
  storageClassName: nfs-client
  accessModes:
    - ReadWriteOnce
  resources:
    requests:
      storage: 1000M
---
apiVersion: v1
kind: PersistentVolumeClaim
metadata:
  name: benchmark-pvc
spec:
  storageClassName: nfs-client
  accessModes:
    - ReadWriteOnce
  resources:
    requests:
      storage: 1000M
---
apiVersion: v1
kind: PersistentVolumeClaim
metadata:
  name: nodes
spec:
  storageClassName: nfs-client
  accessModes:
    - ReadWriteOnce
  resources:
    requests:
      storage: 1000M
"""


AGENT_NODE_TEMPLATE: str = """apiVersion: v1
kind: Service
metadata:
  name: abci{validator_ix}
  labels:
    run: abci{validator_ix}
spec:
  ports:
  - port: 26656
    protocol: TCP
    name: tcp1
  - port: 26657
    protocol: TCP
    name: tcp2
  selector:
    app: agent-node-{validator_ix}
---
apiVersion: apps/v1
kind: Deployment
metadata:
  name: agent-node-{validator_ix}
spec:
  replicas: 1
  selector:
    matchLabels:
      app: agent-node-{validator_ix}
  template:
    metadata:
      labels:
        app: agent-node-{validator_ix}
    spec:
      imagePullSecrets:
      - name: regcred
      restartPolicy: Always
      containers:
      - name: node{validator_ix}
        image: {tendermint_image_name}:{tendermint_image_version}
        imagePullPolicy: Always
        resources:
          limits:
            memory: "1512Mi"
            cpu: "1"
          requests:
            cpu: "0.05"
            memory: "128Mi"
        ports:
          - containerPort: 26656
          - containerPort: 26657
        workingDir: /tendermint
        env:
          - name: HOSTNAME
            value: "node{validator_ix}"
          - name: ID
            value: "{validator_ix}"
          - name: PROXY_APP
            value: tcp://localhost:26658
          - name: TMHOME
            value: /tendermint/node{validator_ix}
          - name: CREATE_EMPTY_BLOCKS
            value: "true"
          - name: LOG_FILE
            value: "/logs/node_{validator_ix}.txt"
        args: ["run", "--no-reload", "--host=0.0.0.0", "--port=8080"]
        volumeMounts:
          - mountPath: /tm_state
            name: persistent-data-tm
          - mountPath: /logs
            name: persistent-data
          - mountPath: /tendermint
            name: nodes

      - name: aea
        image: {open_aea_image_name}:{valory_app}-{open_aea_image_version}
        imagePullPolicy: Always
        resources:
          limits:
            memory: "1512Mi"
            cpu: "1"
          requests:
            cpu: "0.05"
            memory: "128Mi"
        env:
          - name: HOSTNAME
            value: "agent-node-{validator_ix}"
          - name: CLUSTERED
            value: "1"
          - name: LOG_FILE
            value: "/logs/aea_{validator_ix}.txt"
        volumeMounts:
          - mountPath: /logs
            name: persistent-data
          - mountPath: /benchmark
            name: persistent-data-benchmark
          - mountPath: /build
            name: nodes
          - mountPath: /agent_key
            name: agent-key
      volumes:
        - name: agent-key
          secret:
            secretName: agent-validator-{validator_ix}-key
            items:
            - key: private_key
              path: ethereum_private_key.txt
        - name: persistent-data
          persistentVolumeClaim:
            claimName: 'logs-pvc'
        - name: persistent-data-benchmark
          persistentVolumeClaim:
            claimName: 'benchmark-pvc'
        - name: persistent-data-tm
          persistentVolumeClaim:
            claimName: 'tendermint-pvc'
        - name: nodes
          persistentVolumeClaim:
            claimName: 'nodes'
"""

AGENT_SECRET_TEMPLATE: str = """
apiVersion: v1
stringData:
    private_key: '{private_key}'
kind: Secret
metadata:
  annotations:
  name: agent-validator-{validator_ix}-key
type: Opaque
"""



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test base."""

import shutil
import tempfile
from contextlib import suppress
from pathlib import Path
from typing import Tuple

from aea.test_tools.click_testing import CliRunner
from click.testing import Result

from autonomy.cli import cli


class BaseCliTest:
    """Test `autonomy analyse abci` command."""

    t: Path
    cwd: Path
    cli_runner: CliRunner
    cli_options: Tuple[str, ...]

    @classmethod
    def setup(
        cls,
    ) -> None:
        """Setup test."""

        cls.cli_runner = CliRunner()
        cls.cwd = Path.cwd().absolute()
        cls.t = Path(tempfile.mkdtemp())
        cls.t.mkdir(exist_ok=True)

    def run_cli(self, commands: Tuple[str, ...]) -> Result:
        """Run CLI."""
        return self.cli_runner.invoke(cli=cli, args=(*self.cli_options, *commands))

    @classmethod
    def teardown(
        cls,
    ) -> None:
        """Teardown method."""
        with suppress(OSError, FileExistsError, PermissionError):
            shutil.rmtree(str(cls.t))
